"""
    tests.support.pytest.helpers
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    PyTest helpers functions
"""

import logging
import os
import pathlib
import pprint
import shutil
import subprocess
import tempfile
import textwrap
import time
import types
import warnings
from contextlib import contextmanager

import attr
import psutil
import pytest
import requests
from saltfactories.utils import random_string
from saltfactories.utils.tempfiles import temp_file

import salt.utils.platform
import salt.utils.pycrypto
from tests.support.pytest.loader import LoaderModuleMock
from tests.support.runtests import RUNTIME_VARS
from tests.support.sminion import create_sminion

log = logging.getLogger(__name__)


@pytest.helpers.register
def temp_state_file(name, contents, saltenv="base", strip_first_newline=True):
    """
    This helper creates a temporary state file. It should be used as a context manager
    which returns the temporary state file path, and, once out of context, deletes it.

    Can be directly imported and used, or, it can be used as a pytest helper function if
    ``pytest-helpers-namespace`` is installed.

    .. code-block:: python

        import os
        import pytest

        def test_blah():
            with pytest.helpers.temp_state_file("blah.sls") as tpath:
                print(tpath)
                assert os.path.exists(tpath)

            assert not os.path.exists(tpath)

    Depending on the saltenv, it will be created under ``RUNTIME_VARS.TMP_STATE_TREE`` or
    ``RUNTIME_VARS.TMP_PRODENV_STATE_TREE``.

    Args:
        name(str):
            The temporary state file name
        contents(str):
            The contents of the temporary file
        saltenv(str):
            The salt env to use. Either ``base`` or ``prod``
        strip_first_newline(bool):
            Whether to strip the initial first new line char or not.
    """

    if saltenv == "base":
        directory = RUNTIME_VARS.TMP_BASEENV_STATE_TREE
    elif saltenv == "prod":
        directory = RUNTIME_VARS.TMP_PRODENV_STATE_TREE
    else:
        raise RuntimeError(f'"saltenv" can only be "base" or "prod", not "{saltenv}"')
    return temp_file(
        name, contents, directory=directory, strip_first_newline=strip_first_newline
    )


@pytest.helpers.register
def temp_pillar_file(name, contents, saltenv="base", strip_first_newline=True):
    """
    This helper creates a temporary pillar file. It should be used as a context manager
    which returns the temporary pillar file path, and, once out of context, deletes it.

    Can be directly imported and used, or, it can be used as a pytest helper function if
    ``pytest-helpers-namespace`` is installed.

    .. code-block:: python

        import os
        import pytest

        def test_blah():
            with pytest.helpers.temp_pillar_file("blah.sls") as tpath:
                print(tpath)
                assert os.path.exists(tpath)

            assert not os.path.exists(tpath)

    Depending on the saltenv, it will be created under ``RUNTIME_VARS.TMP_PILLAR_TREE`` or
    ``RUNTIME_VARS.TMP_PRODENV_PILLAR_TREE``.

    Args:
        name(str):
            The temporary state file name
        contents(str):
            The contents of the temporary file
        saltenv(str):
            The salt env to use. Either ``base`` or ``prod``
        strip_first_newline(bool):
            Whether to strip the initial first new line char or not.
    """

    if saltenv == "base":
        directory = RUNTIME_VARS.TMP_BASEENV_PILLAR_TREE
    elif saltenv == "prod":
        directory = RUNTIME_VARS.TMP_PRODENV_PILLAR_TREE
    else:
        raise RuntimeError(f'"saltenv" can only be "base" or "prod", not "{saltenv}"')
    return temp_file(
        name, contents, directory=directory, strip_first_newline=strip_first_newline
    )


@pytest.helpers.register
def loader_mock(*args, **kwargs):
    if len(args) > 1:
        loader_modules = args[1]
        warnings.warn(
            "'request' is not longer an accepted argument to 'loader_mock()'. Please"
            " stop passing it.",
            category=DeprecationWarning,
        )
    else:
        loader_modules = args[0]
    return LoaderModuleMock(loader_modules, **kwargs)


@pytest.helpers.register
def salt_loader_module_functions(module):
    if not isinstance(module, types.ModuleType):
        raise RuntimeError(
            "The passed 'module' argument must be an imported "
            "imported module, not {}".format(type(module))
        )
    funcs = {}
    func_alias = getattr(module, "__func_alias__", {})
    virtualname = getattr(module, "__virtualname__")
    for name in dir(module):
        if name.startswith("_"):
            continue
        func = getattr(module, name)
        if getattr(func, "__module__", None) != module.__name__:
            # Not eve defined on the module being processed, carry on
            continue
        if not isinstance(func, types.FunctionType):
            # Not a function? carry on
            continue
        funcname = func_alias.get(func.__name__) or func.__name__
        funcs[f"{virtualname}.{funcname}"] = func
    return funcs


@pytest.helpers.register
def remove_stale_minion_key(master, minion_id):
    key_path = os.path.join(master.config["pki_dir"], "minions", minion_id)
    if os.path.exists(key_path):
        os.unlink(key_path)
    else:
        log.debug("The minion(id=%r) key was not found at %s", minion_id, key_path)


@pytest.helpers.register
def remove_stale_master_key(master):
    keys_path = os.path.join(master.config["pki_dir"], "master")
    for key_name in ("master.pem", "master.pub"):
        key_path = os.path.join(keys_path, key_name)
        if os.path.exists(key_path):
            os.unlink(key_path)
        else:
            log.debug(
                "The master(id=%r) %s key was not found at %s",
                master.id,
                key_name,
                key_path,
            )
    key_path = os.path.join(master.config["pki_dir"], "minion", "minion_master.pub")
    if os.path.exists(key_path):
        os.unlink(key_path)
    else:
        log.debug(
            "The master(id=%r) minion_master.pub key was not found at %s",
            master.id,
            key_path,
        )


@pytest.helpers.register
def remove_stale_proxy_minion_cache_file(proxy_minion, minion_id=None):
    cachefile = os.path.join(
        proxy_minion.config["cachedir"],
        f"dummy-proxy-{minion_id or proxy_minion.id}.cache",
    )
    if os.path.exists(cachefile):
        os.unlink(cachefile)


@attr.s(kw_only=True, slots=True)
class TestGroup:
    sminion = attr.ib(repr=False)
    name = attr.ib()
    gid = attr.ib(default=None)
    members = attr.ib(default=None)
    _delete_group = attr.ib(init=False, repr=False, default=False)

    @sminion.default
    def _default_sminion(self):
        return create_sminion()

    @name.default
    def _default_name(self):
        return random_string("group-", uppercase=False)

    @property
    def info(self):
        return types.SimpleNamespace(**self.sminion.functions.group.info(self.name))

    def __enter__(self):
        group = self.sminion.functions.group.info(self.name)
        if not group:
            ret = self.sminion.functions.group.add(
                self.name, gid=self.gid, non_unique=True
            )
            assert ret
            self._delete_group = True
            log.debug("Created system group: %s", self)
        else:
            log.debug("Reusing existing system group: %s", self)
        if self.members:
            ret = self.sminion.functions.group.members(
                self.name, members_list=self.members
            )
            assert ret
        # Run tests
        return self

    def __exit__(self, *_):
        if self._delete_group:
            try:
                self.sminion.functions.group.delete(self.name)
                log.debug("Deleted system group: %s", self.name)
            except Exception:  # pylint: disable=broad-except
                log.warning(
                    "Failed to delete system group: %s", self.name, exc_info=True
                )


@pytest.helpers.register
@contextmanager
def create_group(name=attr.NOTHING, sminion=attr.NOTHING, gid=None, members=None):
    with TestGroup(sminion=sminion, name=name, gid=gid, members=members) as group:
        yield group


@attr.s(kw_only=True, slots=True)
class TestAccount:
    sminion = attr.ib(repr=False)
    username = attr.ib()
    password = attr.ib()
    hashed_password = attr.ib(repr=False)
    create_group = attr.ib(repr=False, default=False)
    group_name = attr.ib()
    _group = attr.ib(init=True, repr=False)
    _delete_account = attr.ib(init=False, repr=False, default=False)

    @sminion.default
    def _default_sminion(self):
        return create_sminion()

    @username.default
    def _default_username(self):
        return random_string("account-", uppercase=False)

    @password.default
    def _default_password(self):
        return random_string("pwd-", size=8)

    @hashed_password.default
    def _default_hashed_password(self):
        if not salt.utils.platform.is_darwin() and not salt.utils.platform.is_windows():
            return salt.utils.pycrypto.gen_hash(password=self.password)
        return self.password

    @group_name.default
    def _default_group_name(self):
        if self.create_group:
            return f"group-{self.username}"
        return None

    @_group.default
    def _default__group(self):
        if self.group_name:
            return TestGroup(sminion=self.sminion, name=self.group_name)
        return None

    @property
    def info(self):
        return types.SimpleNamespace(**self.sminion.functions.user.info(self.username))

    @property
    def group(self):
        if self._group is None:
            raise RuntimeError(
                "Neither `create_group` nor `group_name` was passed when creating the "
                "account. There's no group attribute in this account instance."
            )
        return self._group

    @group.setter
    def _set_group(self, value):
        self._group = value

    def __enter__(self):
        if not self.sminion.functions.user.info(self.username):
            log.debug("Creating system account: %s", self)
            ret = self.sminion.functions.user.add(self.username)
            assert ret is True
            self._delete_account = True
        if salt.utils.platform.is_windows():
            log.debug("Configuring system account: %s", self)
            ret = self.sminion.functions.user.update(
                self.username, expired=False, password_never_expires=True
            )
            assert ret is True
        if salt.utils.platform.is_darwin() or salt.utils.platform.is_windows():
            password = self.password
        else:
            password = self.hashed_password
        ret = self.sminion.functions.shadow.set_password(self.username, password)
        assert ret is True
        assert self.username in self.sminion.functions.user.list_users()
        if self._group:
            self.group.__enter__()
            self.sminion.functions.group.adduser(self.group.name, self.username)
            if not salt.utils.platform.is_windows():
                # Make this group the primary_group for the user
                self.sminion.functions.user.chgid(self.username, self.group.info.gid)
                assert self.info.gid == self.group.info.gid
        if self._delete_account:
            log.debug("Created system account: %s", self)
        else:
            log.debug("Reusing existing system account: %s", self)
        # Run tests
        return self

    def __exit__(self, *args):
        if self._group:
            try:
                self.sminion.functions.group.deluser(self.group.name, self.username)
                log.debug(
                    "Removed user %r from group %r", self.username, self.group.name
                )
            except Exception:  # pylint: disable=broad-except
                log.warning(
                    "Failed to remove user %r from group %r",
                    self.username,
                    self.group.name,
                    exc_info=True,
                )

            self.group.__exit__(*args)

        if self._delete_account:
            try:
                delete_kwargs = {"force": True}
                if salt.utils.platform.is_windows():
                    delete_kwargs["purge"] = True
                else:
                    delete_kwargs["remove"] = True
                self.sminion.functions.user.delete(self.username, **delete_kwargs)
                log.debug("Deleted system account: %s", self.username)
            except Exception:  # pylint: disable=broad-except
                log.warning(
                    "Failed to delete system account: %s", self.username, exc_info=True
                )

            if self.sminion.functions.group.info(self.username):
                # A group with the same name as the user name still exists.
                # Let's delete it
                try:
                    self.sminion.functions.group.delete(self.username)
                    log.debug(
                        "Deleted system group matching username: %s", self.username
                    )
                except Exception:  # pylint: disable=broad-except
                    log.warning(
                        "Failed to delete system group matching username: %s",
                        self.username,
                        exc_info=True,
                    )


@pytest.helpers.register
@contextmanager
def create_account(
    username=attr.NOTHING,
    password=attr.NOTHING,
    hashed_password=attr.NOTHING,
    group_name=attr.NOTHING,
    create_group=False,
    group=attr.NOTHING,
    sminion=attr.NOTHING,
):
    with TestAccount(
        sminion=sminion,
        username=username,
        password=password,
        hashed_password=hashed_password,
        group_name=group_name,
        create_group=create_group,
        group=group,
    ) as account:
        yield account


@pytest.helpers.register
def shell_test_true():
    if salt.utils.platform.is_windows():
        return "cmd.exe /c exit 0"
    if salt.utils.platform.is_darwin() or salt.utils.platform.is_freebsd():
        return "/usr/bin/true"
    return "/bin/true"


@pytest.helpers.register
def shell_test_false():
    if salt.utils.platform.is_windows():
        return "cmd.exe /c exit 1"
    if salt.utils.platform.is_darwin() or salt.utils.platform.is_freebsd():
        return "/usr/bin/false"
    return "/bin/false"


@attr.s(kw_only=True, frozen=True)
class FakeSaltExtension:
    tmp_path_factory = attr.ib(repr=False)
    name = attr.ib()
    pkgname = attr.ib(init=False)
    srcdir = attr.ib(init=False)

    @srcdir.default
    def _srcdir(self):
        return self.tmp_path_factory.mktemp("src", numbered=True)

    @pkgname.default
    def _pkgname(self):
        replace_chars = ("-", " ")
        name = self.name
        for char in replace_chars:
            name = name.replace(char, "_")
        return name

    def __attrs_post_init__(self):
        self._laydown_files()

    def _laydown_files(self):
        if not self.srcdir.exists():
            self.srcdir.mkdir()
        setup_py = self.srcdir.joinpath("setup.py")
        if not setup_py.exists():
            setup_py.write_text(
                textwrap.dedent(
                    """\
            import setuptools

            if __name__ == "__main__":
                setuptools.setup()
            """
                )
            )
        setup_cfg = self.srcdir.joinpath("setup.cfg")
        if not setup_cfg.exists():
            setup_cfg.write_text(
                textwrap.dedent(
                    """\
            [metadata]
            name = {0}
            version = 1.0
            description = Salt Extension Test
            author = Pedro
            author_email = pedro@algarvio.me
            keywords = salt-extension
            url = http://saltproject.io
            license = Apache Software License 2.0
            classifiers =
                Programming Language :: Python
                Programming Language :: Cython
                Programming Language :: Python :: 3
                Programming Language :: Python :: 3 :: Only
                Development Status :: 4 - Beta
                Intended Audience :: Developers
                License :: OSI Approved :: Apache Software License
            platforms = any

            [options]
            zip_safe = False
            include_package_data = True
            packages = find:
            python_requires = >= 3.5
            setup_requires =
              wheel
              setuptools>=50.3.2
            install_requires =
              distro

            [options.entry_points]
            salt.loader=
              module_dirs = {1}
              runner_dirs = {1}.loader:get_runner_dirs
              states_dirs = {1}.loader:get_state_dirs
              wheel_dirs = {1}.loader:get_new_style_entry_points
            """.format(
                        self.name, self.pkgname
                    )
                )
            )

        extension_package_dir = self.srcdir / self.pkgname
        if not extension_package_dir.exists():
            extension_package_dir.mkdir()
            extension_package_dir.joinpath("__init__.py").write_text("")
            extension_package_dir.joinpath("loader.py").write_text(
                textwrap.dedent(
                    """\
            import pathlib

            PKG_ROOT = pathlib.Path(__file__).resolve().parent

            def get_module_dirs():
                return [str(PKG_ROOT / "modules")]

            def get_runner_dirs():
                return [str(PKG_ROOT / "runners1"), str(PKG_ROOT / "runners2")]

            def get_state_dirs():
                yield str(PKG_ROOT / "states1")

            def get_new_style_entry_points():
                return {"wheel": [str(PKG_ROOT / "the_wheel_modules")]}
            """
                )
            )

            runners1_dir = extension_package_dir / "runners1"
            runners1_dir.mkdir()
            runners1_dir.joinpath("__init__.py").write_text("")
            runners1_dir.joinpath("foobar1.py").write_text(
                textwrap.dedent(
                    """\
            __virtualname__ = "foobar"

            def __virtual__():
                return True

            def echo1(string):
                return string
            """
                )
            )

            runners2_dir = extension_package_dir / "runners2"
            runners2_dir.mkdir()
            runners2_dir.joinpath("__init__.py").write_text("")
            runners2_dir.joinpath("foobar2.py").write_text(
                textwrap.dedent(
                    """\
            __virtualname__ = "foobar"

            def __virtual__():
                return True

            def echo2(string):
                return string
            """
                )
            )

            modules_dir = extension_package_dir / "modules"
            modules_dir.mkdir()
            modules_dir.joinpath("__init__.py").write_text("")
            modules_dir.joinpath("foobar1.py").write_text(
                textwrap.dedent(
                    """\
            __virtualname__ = "foobar"

            def __virtual__():
                return True

            def echo1(string):
                return string
            """
                )
            )
            modules_dir.joinpath("foobar2.py").write_text(
                textwrap.dedent(
                    """\
            __virtualname__ = "foobar"

            def __virtual__():
                return True

            def echo2(string):
                return string
            """
                )
            )

            wheel_dir = extension_package_dir / "the_wheel_modules"
            wheel_dir.mkdir()
            wheel_dir.joinpath("__init__.py").write_text("")
            wheel_dir.joinpath("foobar1.py").write_text(
                textwrap.dedent(
                    """\
            __virtualname__ = "foobar"

            def __virtual__():
                return True

            def echo1(string):
                return string
            """
                )
            )
            wheel_dir.joinpath("foobar2.py").write_text(
                textwrap.dedent(
                    """\
            __virtualname__ = "foobar"

            def __virtual__():
                return True

            def echo2(string):
                return string
            """
                )
            )

            states_dir = extension_package_dir / "states1"
            states_dir.mkdir()
            states_dir.joinpath("__init__.py").write_text("")
            states_dir.joinpath("foobar1.py").write_text(
                textwrap.dedent(
                    """\
            __virtualname__ = "foobar"

            def __virtual__():
                return True

            def echoed(string):
                ret = {"name": name, "changes": {}, "result": True, "comment": string}
                return ret
            """
                )
            )

            utils_dir = extension_package_dir / "utils"
            utils_dir.mkdir()
            utils_dir.joinpath("__init__.py").write_text("")
            utils_dir.joinpath("foobar1.py").write_text(
                textwrap.dedent(
                    """\
            __virtualname__ = "foobar"

            def __virtual__():
                return True

            def echo(string):
                return string
            """
                )
            )

    def __enter__(self):
        self._laydown_files()
        return self

    def __exit__(self, *_):
        shutil.rmtree(str(self.srcdir), ignore_errors=True)


class EntropyGenerator:
    max_minutes = 5
    minimum_entropy = 800

    def __init__(self, max_minutes=None, minimum_entropy=None, skip=None):
        if max_minutes is not None:
            self.max_minutes = max_minutes
        if minimum_entropy is not None:
            self.minimum_entropy = minimum_entropy
        if skip is None:
            skip = True
        self.skip = skip
        self.current_entropy = 0

    def generate_entropy(self):
        max_time = self.max_minutes * 60
        kernel_entropy_file = pathlib.Path("/proc/sys/kernel/random/entropy_avail")
        kernel_poolsize_file = pathlib.Path("/proc/sys/kernel/random/poolsize")
        if not kernel_entropy_file.exists():
            log.info("The '%s' file is not available", kernel_entropy_file)
            return

        self.current_entropy = int(
            kernel_entropy_file.read_text(encoding="utf-8").strip()
        )
        log.info("Available Entropy: %s", self.current_entropy)

        if not kernel_poolsize_file.exists():
            log.info("The '%s' file is not available", kernel_poolsize_file)
        else:
            self.current_poolsize = int(
                kernel_poolsize_file.read_text(encoding="utf-8").strip()
            )
            log.info("Entropy Poolsize: %s", self.current_poolsize)
            # Account for smaller poolsizes using BLAKE2s
            if self.current_poolsize == 256:
                self.minimum_entropy = 192

        if self.current_entropy >= self.minimum_entropy:
            return

        rngd = shutil.which("rngd")
        openssl = shutil.which("openssl")
        timeout = time.time() + max_time
        if rngd:
            log.info("Using rngd to generate entropy")
            while True:
                if time.time() >= timeout:
                    message = (
                        "Skipping test as generating entropy took more than {} minutes. "
                        "Current entropy value {}".format(
                            self.max_minutes, self.current_entropy
                        )
                    )
                    if self.skip:
                        raise pytest.skip.Exception(message, _use_item_location=True)
                    raise pytest.fail(message)
                subprocess.run([rngd, "-r", "/dev/urandom"], shell=False, check=True)
                self.current_entropy = int(
                    kernel_entropy_file.read_text(encoding="utf-8").strip()
                )
                log.info("Available Entropy: %s", self.current_entropy)
                if self.current_entropy >= self.minimum_entropy:
                    break
        elif openssl:
            log.info("Using openssl to generate entropy")
            while True:
                if time.time() >= timeout:
                    message = (
                        "Skipping test as generating entropy took more than {} minutes. "
                        "Current entropy value {}".format(
                            self.max_minutes, self.current_entropy
                        )
                    )
                    if self.skip:
                        raise pytest.skip.Exception(message, _use_item_location=True)
                    raise pytest.fail(message)

                target_file = tempfile.NamedTemporaryFile(
                    delete=False, suffix="sample.txt"
                )
                target_file.close()
                subprocess.run(
                    [
                        openssl,
                        "rand",
                        "-out",
                        target_file.name,
                        "-base64",
                        str(int(2**30 * 3 / 4)),  # 1GB
                    ],
                    shell=False,
                    check=True,
                )
                os.unlink(target_file.name)
                self.current_entropy = int(
                    kernel_entropy_file.read_text(encoding="utf-8").strip()
                )
                log.info("Available Entropy: %s", self.current_entropy)
                if self.current_entropy >= self.minimum_entropy:
                    break
        else:
            message = (
                "Skipping test as there's not enough entropy({}) to continue and "
                "neither 'rgn-tools' nor 'openssl' is available on the system.".format(
                    self.current_entropy
                )
            )
            if self.skip:
                raise pytest.skip.Exception(message, _use_item_location=True)
            raise pytest.fail(message)

    def __enter__(self):
        self.generate_entropy()
        return self

    def __exit__(self, *_):
        pass


@pytest.helpers.register
@contextmanager
def change_cwd(path):
    """
    Context manager helper to change CWD for a with code block and restore
    it at the end
    """
    old_cwd = os.getcwd()
    try:
        os.chdir(path)
        # Do stuff
        yield
    finally:
        # Restore Old CWD
        os.chdir(old_cwd)


@contextmanager
def reap_stray_processes(pid: int = os.getpid()):

    try:
        pre_children = psutil.Process(pid).children(recursive=True)
        # Do stuff
        yield
    finally:
        post_children = psutil.Process(pid).children(recursive=True)

    children = []
    for process in post_children:
        if process in pre_children:
            # Process existed before entering the context
            continue
        if not psutil.pid_exists(process.pid):
            # Process just died
            continue
        # This process is alive and was not running before entering the context
        children.append(process)

    if not children:
        log.info("No astray processes found")
        return

    def on_terminate(proc):
        log.debug("Process %s terminated with exit code %s", proc, proc.returncode)

    if children:
        # Reverse the order, sublings first, parents after
        children.reverse()
        log.warning(
            "Test suite left %d astray processes running. Killing those processes:\n%s",
            len(children),
            pprint.pformat(children),
        )

        _, alive = psutil.wait_procs(children, timeout=3, callback=on_terminate)
        for child in alive:
            try:
                child.kill()
            except psutil.NoSuchProcess:
                continue

        _, alive = psutil.wait_procs(alive, timeout=3, callback=on_terminate)
        if alive:
            # Give up
            for child in alive:
                log.warning(
                    "Process %s survived SIGKILL, giving up:\n%s",
                    child,
                    pprint.pformat(child.as_dict()),
                )


@pytest.helpers.register
def download_file(url, dest, auth=None):
    # NOTE the stream=True parameter below
    with requests.get(url, stream=True, auth=auth, timeout=60) as r:
        r.raise_for_status()
        with salt.utils.files.fopen(dest, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                if chunk:
                    f.write(chunk)
    return dest


# Only allow star importing the functions defined in this module
__all__ = [
    name
    for (name, func) in locals().items()
    if getattr(func, "__module__", None) == __name__
]



"""
Unit tests for the docker state
"""

import pytest

import salt.modules.dockermod as docker_mod
import salt.states.docker_volume as docker_state
from tests.support.mock import Mock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        docker_mod: {"__context__": {"docker.docker_version": ""}},
        docker_state: {"__opts__": {"test": False}},
    }


def test_present():
    """
    Test docker_volume.present
    """
    volumes = []
    default_driver = "dummy_default"

    def create_volume(name, driver=None, driver_opts=None):
        for v in volumes:
            assert v["Name"] != name
        if driver is None:
            driver = default_driver
        new = {"Name": name, "Driver": driver}
        volumes.append(new)
        return new

    def remove_volume(name):
        removed = [v for v in volumes if v["Name"] == name]
        assert 1 == len(removed)
        volumes.remove(removed[0])
        return removed[0]

    docker_create_volume = Mock(side_effect=create_volume)
    __salt__ = {
        "docker.create_volume": docker_create_volume,
        "docker.volumes": Mock(return_value={"Volumes": volumes}),
        "docker.remove_volume": Mock(side_effect=remove_volume),
    }
    with patch.dict(docker_state.__dict__, {"__salt__": __salt__}):
        ret = docker_state.present("volume_foo")
        docker_create_volume.assert_called_with(
            "volume_foo", driver=None, driver_opts=None
        )
        assert ret == {
            "name": "volume_foo",
            "comment": "",
            "changes": {"created": {"Driver": default_driver, "Name": "volume_foo"}},
            "result": True,
        }
        assert len(volumes) == 1
        assert volumes[0]["Name"] == "volume_foo"
        assert volumes[0]["Driver"] is default_driver

        # run it again with the same arguments
        orig_volumes = [volumes[0].copy()]
        ret = docker_state.present("volume_foo")
        assert ret == {
            "name": "volume_foo",
            "comment": "Volume 'volume_foo' already exists.",
            "changes": {},
            "result": True,
        }
        assert orig_volumes == volumes

        # run it again with a different driver but don't force
        ret = docker_state.present("volume_foo", driver="local")
        assert ret == {
            "name": "volume_foo",
            "comment": (
                "Driver for existing volume 'volume_foo'"
                " ('dummy_default') does not match specified"
                " driver ('local') and force is False"
            ),
            "changes": {},
            "result": False,
        }
        assert orig_volumes == volumes

        # run it again with a different driver and force
        ret = docker_state.present("volume_foo", driver="local", force=True)
        assert ret == {
            "name": "volume_foo",
            "comment": "",
            "changes": {
                "removed": {"Driver": default_driver, "Name": "volume_foo"},
                "created": {"Driver": "local", "Name": "volume_foo"},
            },
            "result": True,
        }
        mod_orig_volumes = [orig_volumes[0].copy()]
        mod_orig_volumes[0]["Driver"] = "local"
        assert mod_orig_volumes == volumes


def test_present_with_another_driver():
    """
    Test docker_volume.present
    """
    docker_create_volume = Mock(return_value="created")
    docker_remove_volume = Mock(return_value="removed")
    __salt__ = {
        "docker.create_volume": docker_create_volume,
        "docker.remove_volume": docker_remove_volume,
        "docker.volumes": Mock(
            return_value={"Volumes": [{"Name": "volume_foo", "Driver": "foo"}]}
        ),
    }
    with patch.dict(docker_state.__dict__, {"__salt__": __salt__}):
        ret = docker_state.present(
            "volume_foo",
            driver="bar",
            force=True,
        )
    docker_remove_volume.assert_called_with("volume_foo")
    docker_create_volume.assert_called_with(
        "volume_foo", driver="bar", driver_opts=None
    )
    assert ret == {
        "name": "volume_foo",
        "comment": "",
        "changes": {"created": "created", "removed": "removed"},
        "result": True,
    }


def test_present_wo_existing_volumes():
    """
    Test docker_volume.present without existing volumes.
    """
    docker_create_volume = Mock(return_value="created")
    docker_remove_volume = Mock(return_value="removed")
    __salt__ = {
        "docker.create_volume": docker_create_volume,
        "docker.remove_volume": docker_remove_volume,
        "docker.volumes": Mock(return_value={"Volumes": None}),
    }
    with patch.dict(docker_state.__dict__, {"__salt__": __salt__}):
        ret = docker_state.present(
            "volume_foo",
            driver="bar",
            force=True,
        )
    docker_create_volume.assert_called_with(
        "volume_foo", driver="bar", driver_opts=None
    )
    assert ret == {
        "name": "volume_foo",
        "comment": "",
        "changes": {"created": "created"},
        "result": True,
    }


def test_absent():
    """
    Test docker_volume.absent
    """
    docker_remove_volume = Mock(return_value="removed")
    __salt__ = {
        "docker.remove_volume": docker_remove_volume,
        "docker.volumes": Mock(return_value={"Volumes": [{"Name": "volume_foo"}]}),
    }
    with patch.dict(docker_state.__dict__, {"__salt__": __salt__}):
        ret = docker_state.absent(
            "volume_foo",
        )
    docker_remove_volume.assert_called_with("volume_foo")
    assert ret == {
        "name": "volume_foo",
        "comment": "",
        "changes": {"removed": "removed"},
        "result": True,
    }



"""
Azure (ARM) Compute State Module

.. versionadded:: 2019.2.0

.. warning::

    This cloud provider will be removed from Salt in version 3007 in favor of
    the `saltext.azurerm Salt Extension
    <https://github.com/salt-extensions/saltext-azurerm>`_

:maintainer: <devops@eitr.tech>
:maturity: new
:depends:
    * `azure <https://pypi.python.org/pypi/azure>`_ >= 2.0.0
    * `azure-common <https://pypi.python.org/pypi/azure-common>`_ >= 1.1.8
    * `azure-mgmt <https://pypi.python.org/pypi/azure-mgmt>`_ >= 1.0.0
    * `azure-mgmt-compute <https://pypi.python.org/pypi/azure-mgmt-compute>`_ >= 1.0.0
    * `azure-mgmt-network <https://pypi.python.org/pypi/azure-mgmt-network>`_ >= 1.7.1
    * `azure-mgmt-resource <https://pypi.python.org/pypi/azure-mgmt-resource>`_ >= 1.1.0
    * `azure-mgmt-storage <https://pypi.python.org/pypi/azure-mgmt-storage>`_ >= 1.0.0
    * `azure-mgmt-web <https://pypi.python.org/pypi/azure-mgmt-web>`_ >= 0.32.0
    * `azure-storage <https://pypi.python.org/pypi/azure-storage>`_ >= 0.34.3
    * `msrestazure <https://pypi.python.org/pypi/msrestazure>`_ >= 0.4.21
:platform: linux

:configuration: This module requires Azure Resource Manager credentials to be passed as a dictionary of
    keyword arguments to the ``connection_auth`` parameter in order to work properly. Since the authentication
    parameters are sensitive, it's recommended to pass them to the states via pillar.

    Required provider parameters:

    if using username and password:
      * ``subscription_id``
      * ``username``
      * ``password``

    if using a service principal:
      * ``subscription_id``
      * ``tenant``
      * ``client_id``
      * ``secret``

    Optional provider parameters:

    **cloud_environment**: Used to point the cloud driver to different API endpoints, such as Azure GovCloud. Possible values:
      * ``AZURE_PUBLIC_CLOUD`` (default)
      * ``AZURE_CHINA_CLOUD``
      * ``AZURE_US_GOV_CLOUD``
      * ``AZURE_GERMAN_CLOUD``

    Example Pillar for Azure Resource Manager authentication:

    .. code-block:: yaml

        azurearm:
            user_pass_auth:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                username: fletch
                password: 123pass
            mysubscription:
                subscription_id: 3287abc8-f98a-c678-3bde-326766fd3617
                tenant: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                client_id: ABCDEFAB-1234-ABCD-1234-ABCDEFABCDEF
                secret: XXXXXXXXXXXXXXXXXXXXXXXX
                cloud_environment: AZURE_PUBLIC_CLOUD

    Example states using Azure Resource Manager authentication:

    .. code-block:: jinja

        {% set profile = salt['pillar.get']('azurearm:mysubscription') %}
        Ensure availability set exists:
            azurearm_compute.availability_set_present:
                - name: my_avail_set
                - resource_group: my_rg
                - virtual_machines:
                    - my_vm1
                    - my_vm2
                - tags:
                    how_awesome: very
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}

        Ensure availability set is absent:
            azurearm_compute.availability_set_absent:
                - name: other_avail_set
                - resource_group: my_rg
                - connection_auth: {{ profile }}

"""

# Python libs

import logging
from functools import wraps

import salt.utils.azurearm

__virtualname__ = "azurearm_compute"

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only make this state available if the azurearm_compute module is available.
    """
    if "azurearm_compute.availability_set_create_or_update" in __salt__:
        return __virtualname__
    return (False, "azurearm module could not be loaded")


def _deprecation_message(function):
    """
    Decorator wrapper to warn about azurearm deprecation
    """

    @wraps(function)
    def wrapped(*args, **kwargs):
        salt.utils.versions.warn_until(
            "Chlorine",
            "The 'azurearm' functionality in Salt has been deprecated and its "
            "functionality will be removed in version 3007 in favor of the "
            "saltext.azurerm Salt Extension. "
            "(https://github.com/salt-extensions/saltext-azurerm)",
            category=FutureWarning,
        )
        ret = function(*args, **salt.utils.args.clean_kwargs(**kwargs))
        return ret

    return wrapped


@_deprecation_message
def availability_set_present(
    name,
    resource_group,
    tags=None,
    platform_update_domain_count=None,
    platform_fault_domain_count=None,
    virtual_machines=None,
    sku=None,
    connection_auth=None,
    **kwargs,
):
    """
    .. versionadded:: 2019.2.0

    Ensure an availability set exists.

    :param name:
        Name of the availability set.

    :param resource_group:
        The resource group assigned to the availability set.

    :param tags:
        A dictionary of strings can be passed as tag metadata to the availability set object.

    :param platform_update_domain_count:
        An optional parameter which indicates groups of virtual machines and underlying physical hardware that can be
        rebooted at the same time.

    :param platform_fault_domain_count:
        An optional parameter which defines the group of virtual machines that share a common power source and network
        switch.

    :param virtual_machines:
        A list of names of existing virtual machines to be included in the availability set.

    :param sku:
        The availability set SKU, which specifies whether the availability set is managed or not. Possible values are
        'Aligned' or 'Classic'. An 'Aligned' availability set is managed, 'Classic' is not.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.

    Example usage:

    .. code-block:: yaml

        Ensure availability set exists:
            azurearm_compute.availability_set_present:
                - name: aset1
                - resource_group: group1
                - platform_update_domain_count: 5
                - platform_fault_domain_count: 3
                - sku: aligned
                - tags:
                    contact_name: Elmer Fudd Gantry
                - connection_auth: {{ profile }}
                - require:
                  - azurearm_resource: Ensure resource group exists

    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    if sku:
        sku = {"name": sku.capitalize()}

    aset = __salt__["azurearm_compute.availability_set_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" not in aset:
        tag_changes = __utils__["dictdiffer.deep_diff"](
            aset.get("tags", {}), tags or {}
        )
        if tag_changes:
            ret["changes"]["tags"] = tag_changes

        if platform_update_domain_count and (
            int(platform_update_domain_count)
            != aset.get("platform_update_domain_count")
        ):
            ret["changes"]["platform_update_domain_count"] = {
                "old": aset.get("platform_update_domain_count"),
                "new": platform_update_domain_count,
            }

        if platform_fault_domain_count and (
            int(platform_fault_domain_count) != aset.get("platform_fault_domain_count")
        ):
            ret["changes"]["platform_fault_domain_count"] = {
                "old": aset.get("platform_fault_domain_count"),
                "new": platform_fault_domain_count,
            }

        if sku and (sku["name"] != aset.get("sku", {}).get("name")):
            ret["changes"]["sku"] = {"old": aset.get("sku"), "new": sku}

        if virtual_machines:
            if not isinstance(virtual_machines, list):
                ret["comment"] = "Virtual machines must be supplied as a list!"
                return ret
            aset_vms = aset.get("virtual_machines", [])
            remote_vms = sorted(
                vm["id"].split("/")[-1].lower() for vm in aset_vms if "id" in aset_vms
            )
            local_vms = sorted(vm.lower() for vm in virtual_machines or [])
            if local_vms != remote_vms:
                ret["changes"]["virtual_machines"] = {
                    "old": aset_vms,
                    "new": virtual_machines,
                }

        if not ret["changes"]:
            ret["result"] = True
            ret["comment"] = f"Availability set {name} is already present."
            return ret

        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"Availability set {name} would be updated."
            return ret

    else:
        ret["changes"] = {
            "old": {},
            "new": {
                "name": name,
                "virtual_machines": virtual_machines,
                "platform_update_domain_count": platform_update_domain_count,
                "platform_fault_domain_count": platform_fault_domain_count,
                "sku": sku,
                "tags": tags,
            },
        }

    if __opts__["test"]:
        ret["comment"] = f"Availability set {name} would be created."
        ret["result"] = None
        return ret

    aset_kwargs = kwargs.copy()
    aset_kwargs.update(connection_auth)

    aset = __salt__["azurearm_compute.availability_set_create_or_update"](
        name=name,
        resource_group=resource_group,
        virtual_machines=virtual_machines,
        platform_update_domain_count=platform_update_domain_count,
        platform_fault_domain_count=platform_fault_domain_count,
        sku=sku,
        tags=tags,
        **aset_kwargs,
    )

    if "error" not in aset:
        ret["result"] = True
        ret["comment"] = f"Availability set {name} has been created."
        return ret

    ret["comment"] = "Failed to create availability set {}! ({})".format(
        name, aset.get("error")
    )
    return ret


@_deprecation_message
def availability_set_absent(name, resource_group, connection_auth=None):
    """
    .. versionadded:: 2019.2.0

    Ensure an availability set does not exist in a resource group.

    :param name:
        Name of the availability set.

    :param resource_group:
        Name of the resource group containing the availability set.

    :param connection_auth:
        A dict with subscription and authentication parameters to be used in connecting to the
        Azure Resource Manager API.
    """
    ret = {"name": name, "result": False, "comment": "", "changes": {}}

    if not isinstance(connection_auth, dict):
        ret["comment"] = (
            "Connection information must be specified via connection_auth dictionary!"
        )
        return ret

    aset = __salt__["azurearm_compute.availability_set_get"](
        name, resource_group, azurearm_log_level="info", **connection_auth
    )

    if "error" in aset:
        ret["result"] = True
        ret["comment"] = f"Availability set {name} was not found."
        return ret

    elif __opts__["test"]:
        ret["comment"] = f"Availability set {name} would be deleted."
        ret["result"] = None
        ret["changes"] = {
            "old": aset,
            "new": {},
        }
        return ret

    deleted = __salt__["azurearm_compute.availability_set_delete"](
        name, resource_group, **connection_auth
    )

    if deleted:
        ret["result"] = True
        ret["comment"] = f"Availability set {name} has been deleted."
        ret["changes"] = {"old": aset, "new": {}}
        return ret

    ret["comment"] = f"Failed to delete availability set {name}!"
    return ret



import pytest

from salt.returners import postgres
from tests.support.mock import patch


def test_returner_with_bytes():
    ret = {
        "success": True,
        "return": b"bytes",
        "retcode": 0,
        "jid": "20221101172203459989",
        "fun": "file.read",
        "fun_args": ["/fake/path", {"binary": True}],
        "id": "minion-1",
    }
    with patch.object(postgres, "_get_serv"):
        try:
            postgres.returner(ret)
        except TypeError:
            pytest.fail("Data not decoded properly")


def test_save_load_with_bytes():
    load = {
        "return": b"bytes",
        "jid": "20221101172203459989",
    }
    decoded_load = {
        "return": "bytes",
        "jid": "20221101172203459989",
    }
    with patch.object(postgres, "_get_serv"):
        try:
            postgres.save_load(load["jid"], load)
        except TypeError:
            pytest.fail("Data not decoded properly")



"""
    :codeauthor: Andrew Colin Kissa <andrew@topdog.za.net>
"""

import pytest

import salt.states.postgres_initdb as postgres_initdb
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {postgres_initdb: {}}


def test_present_existing():
    """
    Test existing data directory handled correctly
    """
    name = "/var/lib/psql/data"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_true = MagicMock(return_value=True)
    with patch.dict(postgres_initdb.__salt__, {"postgres.datadir_exists": mock_true}):
        _comt = f"Postgres data directory {name} is already present"
        ret.update({"comment": _comt, "result": True})
        assert postgres_initdb.present(name) == ret


def test_present_non_existing_pass():
    """
    Test non existing data directory ok
    """
    name = "/var/lib/psql/data"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_true = MagicMock(return_value=True)
    mock_false = MagicMock(return_value=False)
    with patch.dict(
        postgres_initdb.__salt__,
        {"postgres.datadir_exists": mock_false, "postgres.datadir_init": mock_true},
    ):
        with patch.dict(postgres_initdb.__opts__, {"test": True}):
            _comt = f"Postgres data directory {name} is set to be initialized"
            ret.update({"comment": _comt, "result": None})
            assert postgres_initdb.present(name) == ret

        with patch.dict(postgres_initdb.__opts__, {"test": False}):
            _comt = f"Postgres data directory {name} has been initialized"
            _changes = {name: "Present"}
            ret.update({"comment": _comt, "result": True, "changes": _changes})
            assert postgres_initdb.present(name) == ret


def test_present_non_existing_fail():
    """
    Test non existing data directory fail
    """
    name = "/var/lib/psql/data"
    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    mock_false = MagicMock(return_value=False)
    with patch.dict(
        postgres_initdb.__salt__,
        {"postgres.datadir_exists": mock_false, "postgres.datadir_init": mock_false},
    ):
        with patch.dict(postgres_initdb.__opts__, {"test": False}):
            _comt = f"Postgres data directory {name} initialization failed"
            ret.update({"comment": _comt, "result": False})
            assert postgres_initdb.present(name) == ret



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)

    =============
    Class Mix-Ins
    =============

    Some reusable class Mixins
"""

import atexit
import copy
import functools
import logging
import multiprocessing
import os
import pprint
import queue
import subprocess
import tempfile
import time
import xml.etree.ElementTree as etree

from saltfactories.utils import random_string

import salt.config
import salt.exceptions
import salt.utils.event
import salt.utils.files
import salt.utils.functools
import salt.utils.path
import salt.utils.process
import salt.utils.stringutils
import salt.utils.yaml
import salt.version
from salt.utils.immutabletypes import freeze
from salt.utils.verify import verify_env
from tests.support.paths import CODE_DIR
from tests.support.pytest.loader import LoaderModuleMock
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


class CheckShellBinaryNameAndVersionMixin:
    """
    Simple class mix-in to subclass in companion to :class:`ShellCase<tests.support.case.ShellCase>` which
    adds a test case to verify proper version report from Salt's CLI tools.
    """

    _call_binary_ = None
    _call_binary_expected_version_ = None

    def test_version_includes_binary_name(self):
        if getattr(self, "_call_binary_", None) is None:
            self.skipTest("'_call_binary_' not defined.")

        if self._call_binary_expected_version_ is None:
            # Late import
            self._call_binary_expected_version_ = salt.version.__version__

        out = "\n".join(self.run_script(self._call_binary_, "--version"))
        # Assert that the binary name is in the output
        try:
            self.assertIn(self._call_binary_, out)
        except AssertionError:
            # We might have generated the CLI scripts in which case we replace '-' with '_'
            alternate_binary_name = self._call_binary_.replace("-", "_")
            errmsg = "Neither '{}' or '{}' were found as part of the binary name in:\n'{}'".format(
                self._call_binary_, alternate_binary_name, out
            )
            self.assertIn(alternate_binary_name, out, msg=errmsg)

        # Assert that the version is in the output
        self.assertIn(self._call_binary_expected_version_, out)


class AdaptedConfigurationTestCaseMixin:

    __slots__ = ()

    @staticmethod
    def get_temp_config(config_for, **config_overrides):
        rootdir = config_overrides.get(
            "root_dir", tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        )
        if not os.path.exists(rootdir):
            os.makedirs(rootdir)
        conf_dir = config_overrides.pop("conf_dir", os.path.join(rootdir, "conf"))
        for key in ("cachedir", "pki_dir", "sock_dir"):
            if key not in config_overrides:
                config_overrides[key] = key
        if "log_file" not in config_overrides:
            config_overrides["log_file"] = f"logs/{config_for}.log"
        if "user" not in config_overrides:
            config_overrides["user"] = RUNTIME_VARS.RUNNING_TESTS_USER
        config_overrides["root_dir"] = rootdir

        cdict = AdaptedConfigurationTestCaseMixin.get_config(
            config_for, from_scratch=True
        )

        if config_for in ("master", "client_config"):
            rdict = salt.config.apply_master_config(config_overrides, cdict)
        if config_for == "minion":
            minion_id = (
                config_overrides.get("id")
                or config_overrides.get("minion_id")
                or cdict.get("id")
                or cdict.get("minion_id")
                or random_string("temp-minion-")
            )
            config_overrides["minion_id"] = config_overrides["id"] = minion_id
            rdict = salt.config.apply_minion_config(
                config_overrides, cdict, cache_minion_id=False, minion_id=minion_id
            )

        verify_env(
            [
                os.path.join(rdict["pki_dir"], "minions"),
                os.path.join(rdict["pki_dir"], "minions_pre"),
                os.path.join(rdict["pki_dir"], "minions_rejected"),
                os.path.join(rdict["pki_dir"], "minions_denied"),
                os.path.join(rdict["cachedir"], "jobs"),
                os.path.join(rdict["cachedir"], "tokens"),
                os.path.join(rdict["root_dir"], "cache", "tokens"),
                os.path.join(rdict["pki_dir"], "accepted"),
                os.path.join(rdict["pki_dir"], "rejected"),
                os.path.join(rdict["pki_dir"], "pending"),
                os.path.dirname(rdict["log_file"]),
                rdict["sock_dir"],
                conf_dir,
            ],
            RUNTIME_VARS.RUNNING_TESTS_USER,
            root_dir=rdict["root_dir"],
        )

        rdict["conf_file"] = os.path.join(conf_dir, config_for)
        with salt.utils.files.fopen(rdict["conf_file"], "w") as wfh:
            salt.utils.yaml.safe_dump(rdict, wfh, default_flow_style=False)
        return rdict

    @staticmethod
    def get_config(config_for, from_scratch=False):
        if from_scratch:
            if config_for in ("master", "syndic_master", "mm_master", "mm_sub_master"):
                return salt.config.master_config(
                    AdaptedConfigurationTestCaseMixin.get_config_file_path(config_for)
                )
            elif config_for in ("minion", "sub_minion"):
                return salt.config.minion_config(
                    AdaptedConfigurationTestCaseMixin.get_config_file_path(config_for),
                    cache_minion_id=False,
                )
            elif config_for in ("syndic",):
                return salt.config.syndic_config(
                    AdaptedConfigurationTestCaseMixin.get_config_file_path(config_for),
                    AdaptedConfigurationTestCaseMixin.get_config_file_path("minion"),
                )
            elif config_for == "client_config":
                return salt.config.client_config(
                    AdaptedConfigurationTestCaseMixin.get_config_file_path("master")
                )

        if config_for not in RUNTIME_VARS.RUNTIME_CONFIGS:
            if config_for in ("master", "syndic_master", "mm_master", "mm_sub_master"):
                RUNTIME_VARS.RUNTIME_CONFIGS[config_for] = freeze(
                    salt.config.master_config(
                        AdaptedConfigurationTestCaseMixin.get_config_file_path(
                            config_for
                        )
                    )
                )
            elif config_for in ("minion", "sub_minion"):
                RUNTIME_VARS.RUNTIME_CONFIGS[config_for] = freeze(
                    salt.config.minion_config(
                        AdaptedConfigurationTestCaseMixin.get_config_file_path(
                            config_for
                        )
                    )
                )
            elif config_for in ("syndic",):
                RUNTIME_VARS.RUNTIME_CONFIGS[config_for] = freeze(
                    salt.config.syndic_config(
                        AdaptedConfigurationTestCaseMixin.get_config_file_path(
                            config_for
                        ),
                        AdaptedConfigurationTestCaseMixin.get_config_file_path(
                            "minion"
                        ),
                    )
                )
            elif config_for == "client_config":
                RUNTIME_VARS.RUNTIME_CONFIGS[config_for] = freeze(
                    salt.config.client_config(
                        AdaptedConfigurationTestCaseMixin.get_config_file_path("master")
                    )
                )
        return RUNTIME_VARS.RUNTIME_CONFIGS[config_for]

    @property
    def config_dir(self):
        return RUNTIME_VARS.TMP_CONF_DIR

    def get_config_dir(self):
        log.warning("Use the config_dir attribute instead of calling get_config_dir()")
        return self.config_dir

    @staticmethod
    def get_config_file_path(filename):
        if filename == "master":
            return os.path.join(RUNTIME_VARS.TMP_CONF_DIR, filename)
        if filename == "minion":
            return os.path.join(RUNTIME_VARS.TMP_MINION_CONF_DIR, filename)
        if filename == "syndic_master":
            return os.path.join(RUNTIME_VARS.TMP_SYNDIC_MASTER_CONF_DIR, "master")
        if filename == "syndic":
            return os.path.join(RUNTIME_VARS.TMP_SYNDIC_MINION_CONF_DIR, "minion")
        if filename == "sub_minion":
            return os.path.join(RUNTIME_VARS.TMP_SUB_MINION_CONF_DIR, "minion")
        if filename == "mm_master":
            return os.path.join(RUNTIME_VARS.TMP_MM_CONF_DIR, "master")
        if filename == "mm_sub_master":
            return os.path.join(RUNTIME_VARS.TMP_MM_SUB_CONF_DIR, "master")
        if filename == "mm_minion":
            return os.path.join(RUNTIME_VARS.TMP_MM_MINION_CONF_DIR, "minion")
        if filename == "mm_sub_minion":
            return os.path.join(RUNTIME_VARS.TMP_MM_SUB_MINION_CONF_DIR, "minion")
        return os.path.join(RUNTIME_VARS.TMP_CONF_DIR, filename)

    @property
    def master_opts(self):
        """
        Return the options used for the master
        """
        return self.get_config("master")

    @property
    def minion_opts(self):
        """
        Return the options used for the minion
        """
        return self.get_config("minion")

    @property
    def sub_minion_opts(self):
        """
        Return the options used for the sub_minion
        """
        return self.get_config("sub_minion")


class SaltClientTestCaseMixin(AdaptedConfigurationTestCaseMixin):
    """
    Mix-in class that provides a ``client`` attribute which returns a Salt
    :class:`LocalClient<salt:salt.client.LocalClient>`.

    .. code-block:: python

        class LocalClientTestCase(TestCase, SaltClientTestCaseMixin):

            def test_check_pub_data(self):
                just_minions = {'minions': ['m1', 'm2']}
                jid_no_minions = {'jid': '1234', 'minions': []}
                valid_pub_data = {'minions': ['m1', 'm2'], 'jid': '1234'}

                self.assertRaises(EauthAuthenticationError,
                                  self.client._check_pub_data, None)
                self.assertDictEqual({},
                    self.client._check_pub_data(just_minions),
                    'Did not handle lack of jid correctly')

                self.assertDictEqual(
                    {},
                    self.client._check_pub_data({'jid': '0'}),
                    'Passing JID of zero is not handled gracefully')
    """

    _salt_client_config_file_name_ = "master"

    @property
    def client(self):
        # Late import
        import salt.client

        if "runtime_client" not in RUNTIME_VARS.RUNTIME_CONFIGS:
            mopts = self.get_config(
                self._salt_client_config_file_name_, from_scratch=True
            )
            RUNTIME_VARS.RUNTIME_CONFIGS["runtime_client"] = (
                salt.client.get_local_client(mopts=mopts)
            )
        return RUNTIME_VARS.RUNTIME_CONFIGS["runtime_client"]


class ShellCaseCommonTestsMixin(CheckShellBinaryNameAndVersionMixin):

    _call_binary_expected_version_ = salt.version.__version__

    def test_salt_with_git_version(self):
        if getattr(self, "_call_binary_", None) is None:
            self.skipTest("'_call_binary_' not defined.")
        from salt.version import SaltStackVersion, __version_info__

        git = salt.utils.path.which("git")
        if not git:
            self.skipTest("The git binary is not available")
        opts = {
            "stdout": subprocess.PIPE,
            "stderr": subprocess.PIPE,
            "cwd": CODE_DIR,
        }
        if not salt.utils.platform.is_windows():
            opts["close_fds"] = True
        # Let's get the output of git describe
        process = subprocess.Popen(
            [git, "describe", "--tags", "--first-parent", "--match", "v[0-9]*"], **opts
        )
        out, err = process.communicate()
        if process.returncode != 0:
            process = subprocess.Popen(
                [git, "describe", "--tags", "--match", "v[0-9]*"], **opts
            )
            out, err = process.communicate()
        if not out:
            self.skipTest(
                "Failed to get the output of 'git describe'. Error: '{}'".format(
                    salt.utils.stringutils.to_str(err)
                )
            )

        parsed_version = SaltStackVersion.parse(out)

        if parsed_version.info < __version_info__:
            self.skipTest(
                "We're likely about to release a new version. This test "
                "would fail. Parsed('{}') < Expected('{}')".format(
                    parsed_version.info, __version_info__
                )
            )
        elif parsed_version.info != __version_info__:
            self.skipTest(
                "In order to get the proper salt version with the "
                "git hash you need to update salt's local git "
                "tags. Something like: 'git fetch --tags' or "
                "'git fetch --tags upstream' if you followed "
                "salt's contribute documentation. The version "
                "string WILL NOT include the git hash."
            )
        out = "\n".join(self.run_script(self._call_binary_, "--version"))
        self.assertIn(parsed_version.string, out)


class _FixLoaderModuleMockMixinMroOrder(type):
    """
    This metaclass will make sure that LoaderModuleMockMixin will always come as the first
    base class in order for LoaderModuleMockMixin.setUp to actually run
    """

    def __new__(mcs, cls_name, cls_bases, cls_dict):
        if cls_name == "LoaderModuleMockMixin":
            return super().__new__(mcs, cls_name, cls_bases, cls_dict)
        bases = list(cls_bases)
        for idx, base in enumerate(bases):
            if base.__name__ == "LoaderModuleMockMixin":
                bases.insert(0, bases.pop(idx))
                break

        # Create the class instance
        instance = super().__new__(mcs, cls_name, tuple(bases), cls_dict)

        # Apply our setUp function decorator
        instance.setUp = LoaderModuleMockMixin.__setup_loader_modules_mocks__(
            instance.setUp
        )
        return instance


class LoaderModuleMockMixin(metaclass=_FixLoaderModuleMockMixinMroOrder):
    """
    This class will setup salt loader dunders.

    Please check `set_up_loader_mocks` above
    """

    # Define our setUp function decorator
    @staticmethod
    def __setup_loader_modules_mocks__(setup_func):
        @functools.wraps(setup_func)
        def wrapper(self):
            loader_modules_configs = self.setup_loader_modules()
            if not isinstance(loader_modules_configs, dict):
                raise RuntimeError(
                    "{}.setup_loader_modules() must return a dictionary where the keys"
                    " are the modules that require loader mocking setup and the values,"
                    " the global module variables for each of the module being mocked."
                    " For example '__salt__', '__opts__', etc.".format(
                        self.__class__.__name__
                    )
                )

            mocker = LoaderModuleMock(loader_modules_configs)
            mocker.start()
            self.addCleanup(mocker.stop)
            return setup_func(self)

        return wrapper

    def setup_loader_modules(self):
        raise NotImplementedError(
            "'{}.setup_loader_modules()' must be implemented".format(
                self.__class__.__name__
            )
        )


class XMLEqualityMixin:
    def assertEqualXML(self, e1, e2):
        if isinstance(e1, bytes):
            e1 = e1.decode("utf-8")
        if isinstance(e2, bytes):
            e2 = e2.decode("utf-8")
        if isinstance(e1, str):
            e1 = etree.XML(e1)
        if isinstance(e2, str):
            e2 = etree.XML(e2)
        if e1.tag != e2.tag:
            return False
        if e1.text != e2.text:
            return False
        if e1.tail != e2.tail:
            return False
        if e1.attrib != e2.attrib:
            return False
        if len(e1) != len(e2):
            return False
        return all(self.assertEqualXML(c1, c2) for c1, c2 in zip(e1, e2))


class SaltReturnAssertsMixin:
    def assertReturnSaltType(self, ret):
        try:
            self.assertTrue(isinstance(ret, dict))
        except AssertionError:
            raise AssertionError(
                f"{type(ret).__name__} is not dict. Salt returned: {ret}"
            )

    def assertReturnNonEmptySaltType(self, ret):
        self.assertReturnSaltType(ret)
        try:
            self.assertNotEqual(ret, {})
        except AssertionError:
            raise AssertionError(
                "{} is equal to {}. Salt returned an empty dictionary."
            )

    def __return_valid_keys(self, keys):
        if isinstance(keys, tuple):
            # If it's a tuple, turn it into a list
            keys = list(keys)
        elif isinstance(keys, str):
            # If it's a string, make it a one item list
            keys = [keys]
        elif not isinstance(keys, list):
            # If we've reached here, it's a bad type passed to keys
            raise RuntimeError("The passed keys need to be a list")
        return keys

    def __getWithinSaltReturn(self, ret, keys):
        self.assertReturnNonEmptySaltType(ret)
        ret_data = []
        for part in ret.values():
            keys = self.__return_valid_keys(keys)
            okeys = keys[:]
            try:
                ret_item = part[okeys.pop(0)]
            except (KeyError, TypeError):
                raise AssertionError(
                    "Could not get ret{} from salt's return: {}".format(
                        "".join([f"['{k}']" for k in keys]), part
                    )
                )
            while okeys:
                try:
                    ret_item = ret_item[okeys.pop(0)]
                except (KeyError, TypeError):
                    raise AssertionError(
                        "Could not get ret{} from salt's return: {}".format(
                            "".join([f"['{k}']" for k in keys]), part
                        )
                    )
            ret_data.append(ret_item)
        return ret_data

    def assertSaltTrueReturn(self, ret):
        try:
            for saltret in self.__getWithinSaltReturn(ret, "result"):
                self.assertTrue(saltret)
        except AssertionError:
            log.info("Salt Full Return:\n%s", pprint.pformat(ret))
            try:
                raise AssertionError(
                    "{result} is not True. Salt Comment:\n{comment}".format(
                        **(next(iter(ret.values())))
                    )
                )
            except (AttributeError, IndexError):
                raise AssertionError(
                    "Failed to get result. Salt Returned:\n{}".format(
                        pprint.pformat(ret)
                    )
                )

    def assertSaltFalseReturn(self, ret):
        try:
            for saltret in self.__getWithinSaltReturn(ret, "result"):
                self.assertFalse(saltret)
        except AssertionError:
            log.info("Salt Full Return:\n%s", pprint.pformat(ret))
            try:
                raise AssertionError(
                    "{result} is not False. Salt Comment:\n{comment}".format(
                        **(next(iter(ret.values())))
                    )
                )
            except (AttributeError, IndexError):
                raise AssertionError(f"Failed to get result. Salt Returned: {ret}")

    def assertSaltNoneReturn(self, ret):
        try:
            for saltret in self.__getWithinSaltReturn(ret, "result"):
                self.assertIsNone(saltret)
        except AssertionError:
            log.info("Salt Full Return:\n%s", pprint.pformat(ret))
            try:
                raise AssertionError(
                    "{result} is not None. Salt Comment:\n{comment}".format(
                        **(next(iter(ret.values())))
                    )
                )
            except (AttributeError, IndexError):
                raise AssertionError(f"Failed to get result. Salt Returned: {ret}")

    def assertInSaltComment(self, in_comment, ret):
        for saltret in self.__getWithinSaltReturn(ret, "comment"):
            self.assertIn(in_comment, saltret)

    def assertNotInSaltComment(self, not_in_comment, ret):
        for saltret in self.__getWithinSaltReturn(ret, "comment"):
            self.assertNotIn(not_in_comment, saltret)

    def assertSaltCommentRegexpMatches(self, ret, pattern):
        return self.assertInSaltReturnRegexpMatches(ret, pattern, "comment")

    def assertInSaltStateWarning(self, in_comment, ret):
        for saltret in self.__getWithinSaltReturn(ret, "warnings"):
            self.assertIn(in_comment, saltret)

    def assertNotInSaltStateWarning(self, not_in_comment, ret):
        for saltret in self.__getWithinSaltReturn(ret, "warnings"):
            self.assertNotIn(not_in_comment, saltret)

    def assertInSaltReturn(self, item_to_check, ret, keys):
        for saltret in self.__getWithinSaltReturn(ret, keys):
            self.assertIn(item_to_check, saltret)

    def assertNotInSaltReturn(self, item_to_check, ret, keys):
        for saltret in self.__getWithinSaltReturn(ret, keys):
            self.assertNotIn(item_to_check, saltret)

    def assertInSaltReturnRegexpMatches(self, ret, pattern, keys=()):
        for saltret in self.__getWithinSaltReturn(ret, keys):
            self.assertRegex(saltret, pattern)

    def assertSaltStateChangesEqual(self, ret, comparison, keys=()):
        keys = ["changes"] + self.__return_valid_keys(keys)
        for saltret in self.__getWithinSaltReturn(ret, keys):
            self.assertEqual(saltret, comparison)

    def assertSaltStateChangesNotEqual(self, ret, comparison, keys=()):
        keys = ["changes"] + self.__return_valid_keys(keys)
        for saltret in self.__getWithinSaltReturn(ret, keys):
            self.assertNotEqual(saltret, comparison)


def _fetch_events(q, opts):
    """
    Collect events and store them
    """

    def _clean_queue():
        log.info("Cleaning queue!")
        while not q.empty():
            queue_item = q.get()
            queue_item.task_done()

    atexit.register(_clean_queue)
    with salt.utils.event.get_event(
        "minion", sock_dir=opts["sock_dir"], opts=opts
    ) as event:

        # Wait for event bus to be connected
        while not event.connect_pull(30):
            time.sleep(1)

        # Notify parent process that the event bus is connected
        q.put("CONNECTED")

        while True:
            try:
                events = event.get_event(full=False)
            except Exception as exc:  # pylint: disable=broad-except
                # This is broad but we'll see all kinds of issues right now
                # if we drop the proc out from under the socket while we're reading
                log.exception("Exception caught while getting events %r", exc)
            q.put(events)


class SaltMinionEventAssertsMixin:
    """
    Asserts to verify that a given event was seen
    """

    @classmethod
    def setUpClass(cls):
        opts = copy.deepcopy(RUNTIME_VARS.RUNTIME_CONFIGS["minion"])
        cls.q = multiprocessing.Queue()
        cls.fetch_proc = salt.utils.process.SignalHandlingProcess(
            target=_fetch_events,
            args=(cls.q, opts),
            name=f"Process-{cls.__name__}-Queue",
        )
        cls.fetch_proc.start()
        # Wait for the event bus to be connected
        msg = cls.q.get(block=True)
        if msg != "CONNECTED":
            # Just in case something very bad happens
            raise RuntimeError("Unexpected message in test's event queue")

    @classmethod
    def tearDownClass(cls):
        cls.fetch_proc.join()
        del cls.q
        del cls.fetch_proc

    def assertMinionEventFired(self, tag):
        # TODO
        raise salt.exceptions.NotImplemented("assertMinionEventFired() not implemented")

    def assertMinionEventReceived(self, desired_event, timeout=5, sleep_time=0.5):
        start = time.time()
        while True:
            try:
                event = self.q.get(False)
            except queue.Empty:
                time.sleep(sleep_time)
                if time.time() - start >= timeout:
                    break
                continue
            if isinstance(event, dict):
                event.pop("_stamp")
            if desired_event == event:
                self.fetch_proc.terminate()
                return True
            if time.time() - start >= timeout:
                break
        self.fetch_proc.terminate()
        raise AssertionError(f"Event {desired_event} was not received by minion")



"""
Tests for the nacl runner
"""

import pytest

import salt.config
import salt.utils.stringutils
from tests.support.mock import patch

pytest.importorskip("nacl.public")
pytest.importorskip("nacl.secret")

import salt.runners.nacl as nacl

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture(scope="module")
def minion_opts():
    return salt.config.minion_config(None)


@pytest.fixture
def test_data():
    unencrypted_data = salt.utils.stringutils.to_bytes("hello")
    return unencrypted_data


def test_keygen(minion_opts):
    """
    Test keygen
    """
    # Store the data
    with patch("salt.runners.nacl.__opts__", minion_opts, create=True):
        ret = nacl.keygen()
    assert "pk" in ret
    assert "sk" in ret


def test_enc(test_data, minion_opts):
    """
    Test keygen
    """
    # Store the data
    with patch("salt.runners.nacl.__opts__", minion_opts, create=True):
        ret = nacl.keygen()
        assert "pk" in ret
        assert "sk" in ret
        pk = ret["pk"]
        sk = ret["sk"]

        # Encrypt with pk
        ret = nacl.enc(
            data=test_data,
            pk=pk,
        )


def test_enc_dec(test_data, minion_opts):
    """
    Store, list, fetch, then flush data
    """
    # Store the data
    with patch("salt.runners.nacl.__opts__", minion_opts, create=True):
        ret = nacl.keygen()
        assert "pk" in ret
        assert "sk" in ret
        pk = ret["pk"]
        sk = ret["sk"]

        # Encrypt with pk
        encrypted_data = nacl.enc(
            data=test_data,
            pk=pk,
        )

        # Decrypt with sk
        ret = nacl.dec(
            data=encrypted_data,
            sk=sk,
        )
        assert test_data == ret


def test_sealedbox_enc_dec(test_data, minion_opts):
    """
    Generate keys, encrypt, then decrypt.
    """
    # Store the data
    with patch("salt.runners.nacl.__opts__", minion_opts, create=True):
        ret = nacl.keygen()
        assert "pk" in ret
        assert "sk" in ret
        pk = ret["pk"]
        sk = ret["sk"]

        # Encrypt with pk
        encrypted_data = nacl.sealedbox_encrypt(
            data=test_data,
            pk=pk,
        )

        # Decrypt with sk
        ret = nacl.sealedbox_decrypt(
            data=encrypted_data,
            sk=sk,
        )
        assert test_data == ret


def test_secretbox_enc_dec(test_data, minion_opts):
    """
    Generate keys, encrypt, then decrypt.
    """
    # Store the data
    with patch("salt.runners.nacl.__opts__", minion_opts, create=True):
        ret = nacl.keygen()
        assert "pk" in ret
        assert "sk" in ret
        pk = ret["pk"]
        sk = ret["sk"]

        # Encrypt with pk
        encrypted_data = nacl.secretbox_encrypt(
            data=test_data,
            sk=sk,
        )

        # Decrypt with sk
        ret = nacl.secretbox_decrypt(
            data=encrypted_data,
            sk=sk,
        )
        assert test_data == ret


def test_enc_dec_no_pk_no_sk(test_data, minion_opts):
    """
    Store, list, fetch, then flush data
    """
    # Store the data
    with patch("salt.runners.nacl.__opts__", minion_opts, create=True):
        ret = nacl.keygen()
        assert "pk" in ret
        assert "sk" in ret
        pk = ret["pk"]
        sk = ret["sk"]

        # Encrypt with pk
        with pytest.raises(Exception, match="no pubkey or pk_file found"):
            ret = nacl.enc(
                data=test_data,
                pk=None,
            )

        encrypted_data = test_data  # dummy data, should get exception
        # Decrypt with sk
        with pytest.raises(Exception, match="no key or sk_file found"):
            ret = nacl.dec(
                data=encrypted_data,
                sk=None,
            )



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>

    Test cases for salt.modules.xapi
"""

import pytest

import salt.modules.xapi_virt as xapi
from tests.support.mock import MagicMock, mock_open, patch


class Mockxapi:
    """
    Mock xapi class
    """

    def __init__(self):
        pass

    class Session:
        """
        Mock Session class
        """

        def __init__(self, xapi_uri):
            pass

        class xenapi:
            """
            Mock xenapi class
            """

            def __init__(self):
                pass

            @staticmethod
            def login_with_password(xapi_login, xapi_password):
                """
                Mock login_with_password method
                """
                return xapi_login, xapi_password

            class session:
                """
                Mock session class
                """

                def __init__(self):
                    pass

                @staticmethod
                def logout():
                    """
                    Mock logout method
                    """
                    return Mockxapi()


@pytest.fixture
def configure_loader_modules():
    return {xapi: {}}


def test_list_domains():
    """
    Test to return a list of domain names on the minion
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        assert xapi.list_domains() == []


def test_vm_info():
    """
    Test to return detailed information about the vms
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(return_value=False)
        with patch.object(xapi, "_get_record_by_label", mock):
            assert xapi.vm_info(True) == {True: False}


def test_vm_state():
    """
    Test to return list of all the vms and their state.
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(return_value={"power_state": "1"})
        with patch.object(xapi, "_get_record_by_label", mock):
            assert xapi.vm_state("salt") == {"salt": "1"}
            assert xapi.vm_state() == {}


def test_get_nics():
    """
    Test to return info about the network interfaces of a named vm
    """
    ret = {"Stack": {"device": "ETH0", "mac": "Stack", "mtu": 1}}
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, {"VIFs": "salt"}])
        with patch.object(xapi, "_get_record_by_label", mock):
            assert not xapi.get_nics("salt")
            mock = MagicMock(return_value={"MAC": "Stack", "device": "ETH0", "MTU": 1})
            with patch.object(xapi, "_get_record", mock):
                assert xapi.get_nics("salt") == ret


def test_get_macs():
    """
    Test to return a list off MAC addresses from the named vm
    """
    mock = MagicMock(side_effect=[None, ["a", "b", "c"]])
    with patch.object(xapi, "get_nics", mock):
        assert xapi.get_macs("salt") is None
        assert xapi.get_macs("salt") == ["a", "b", "c"]


def test_get_disks():
    """
    Test to return the disks of a named vm
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, ["a", "b", "c"]])
        with patch.object(xapi, "_get_label_uuid", mock):
            assert not xapi.get_disks("salt")
            assert xapi.get_disks("salt") == {}


def test_setmem():
    """
    Test to changes the amount of memory allocated to VM.
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, ["a", "b", "c"]])
        with patch.object(xapi, "_get_label_uuid", mock):
            assert not xapi.setmem("salt", "1")
            assert xapi.setmem("salt", "1")
    with patch.object(xapi, "_check_xenapi", MagicMock(return_value=Mockxapi)):
        mock = MagicMock(return_value=True)
        with patch.dict(xapi.__salt__, {"config.option": mock}):
            with patch.object(xapi, "_get_label_uuid", mock):
                assert not xapi.setmem("salt", "1")


def test_setvcpus():
    """
    Test to changes the amount of vcpus allocated to VM.
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, ["a", "b", "c"]])
        with patch.object(xapi, "_get_label_uuid", mock):
            assert not xapi.setvcpus("salt", "1")
            assert xapi.setvcpus("salt", "1")
    with patch.object(xapi, "_check_xenapi", MagicMock(return_value=Mockxapi)):
        mock = MagicMock(return_value=True)
        with patch.dict(xapi.__salt__, {"config.option": mock}):
            with patch.object(xapi, "_get_label_uuid", mock):
                assert not xapi.setvcpus("salt", "1")


def test_vcpu_pin():
    """
    Test to Set which CPUs a VCPU can use.
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, ["a", "b", "c"]])
        with patch.object(xapi, "_get_label_uuid", mock):
            assert not xapi.vcpu_pin("salt", "1", "2")
            assert xapi.vcpu_pin("salt", "1", "2")
    with patch.object(xapi, "_check_xenapi", MagicMock(return_value=Mockxapi)):
        mock = MagicMock(return_value=True)
        with patch.dict(xapi.__salt__, {"config.option": mock}):
            with patch.object(xapi, "_get_label_uuid", mock):
                with patch.dict(xapi.__salt__, {"cmd.run": mock}):
                    assert xapi.vcpu_pin("salt", "1", "2")


def test_freemem():
    """
    Test to return an int representing the amount of memory
    that has not been given to virtual machines on this node
    """
    mock = MagicMock(return_value={"free_memory": 1024})
    with patch.object(xapi, "node_info", mock):
        assert xapi.freemem() == 1024


def test_freecpu():
    """
    Test to return an int representing the number
    of unallocated cpus on this hypervisor
    """
    mock = MagicMock(return_value={"free_cpus": 1024})
    with patch.object(xapi, "node_info", mock):
        assert xapi.freecpu() == 1024


def test_full_info():
    """
    Test to return the node_info, vm_info and freemem
    """
    mock = MagicMock(return_value="salt")
    with patch.object(xapi, "node_info", mock):
        mock = MagicMock(return_value="stack")
        with patch.object(xapi, "vm_info", mock):
            assert xapi.full_info() == {"node_info": "salt", "vm_info": "stack"}


def test_shutdown():
    """
    Test to send a soft shutdown signal to the named vm
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, ["a", "b", "c"]])
        with patch.object(xapi, "_get_label_uuid", mock):
            assert not xapi.shutdown("salt")
            assert xapi.shutdown("salt")
    with patch.object(xapi, "_check_xenapi", MagicMock(return_value=Mockxapi)):
        mock = MagicMock(return_value=True)
        with patch.dict(xapi.__salt__, {"config.option": mock}):
            with patch.object(xapi, "_get_label_uuid", mock):
                assert not xapi.shutdown("salt")


def test_pause():
    """
    Test to pause the named vm
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, ["a", "b", "c"]])
        with patch.object(xapi, "_get_label_uuid", mock):
            assert not xapi.pause("salt")
            assert xapi.pause("salt")
    with patch.object(xapi, "_check_xenapi", MagicMock(return_value=Mockxapi)):
        mock = MagicMock(return_value=True)
        with patch.dict(xapi.__salt__, {"config.option": mock}):
            with patch.object(xapi, "_get_label_uuid", mock):
                assert not xapi.pause("salt")


def test_resume():
    """
    Test to resume the named vm
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, ["a", "b", "c"]])
        with patch.object(xapi, "_get_label_uuid", mock):
            assert not xapi.resume("salt")
            assert xapi.resume("salt")
    with patch.object(xapi, "_check_xenapi", MagicMock(return_value=Mockxapi)):
        mock = MagicMock(return_value=True)
        with patch.dict(xapi.__salt__, {"config.option": mock}):
            with patch.object(xapi, "_get_label_uuid", mock):
                assert not xapi.resume("salt")


def test_start():
    """
    Test to reboot a domain via ACPI request
    """
    mock = MagicMock(return_value=True)
    with patch.object(xapi, "start", mock):
        assert xapi.start("salt")


def test_reboot():
    """
    Test to reboot a domain via ACPI request
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, ["a", "b", "c"]])
        with patch.object(xapi, "_get_label_uuid", mock):
            assert not xapi.reboot("salt")
            assert xapi.reboot("salt")
    with patch.object(xapi, "_check_xenapi", MagicMock(return_value=Mockxapi)):
        mock = MagicMock(return_value=True)
        with patch.dict(xapi.__salt__, {"config.option": mock}):
            with patch.object(xapi, "_get_label_uuid", mock):
                assert not xapi.reboot("salt")


def test_reset():
    """
    Test to reset a VM by emulating the
    reset button on a physical machine
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, ["a", "b", "c"]])
        with patch.object(xapi, "_get_label_uuid", mock):
            assert not xapi.reset("salt")
            assert xapi.reset("salt")
    with patch.object(xapi, "_check_xenapi", MagicMock(return_value=Mockxapi)):
        mock = MagicMock(return_value=True)
        with patch.dict(xapi.__salt__, {"config.option": mock}):
            with patch.object(xapi, "_get_label_uuid", mock):
                assert not xapi.reset("salt")


def test_migrate():
    """
    Test to migrates the virtual machine to another hypervisor
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, ["a", "b", "c"]])
        with patch.object(xapi, "_get_label_uuid", mock):
            assert not xapi.migrate("salt", "stack")
            assert xapi.migrate("salt", "stack")
    with patch.object(xapi, "_check_xenapi", MagicMock(return_value=Mockxapi)):
        mock = MagicMock(return_value=True)
        with patch.dict(xapi.__salt__, {"config.option": mock}):
            with patch.object(xapi, "_get_label_uuid", mock):
                assert not xapi.migrate("salt", "stack")


def test_stop():
    """
    Test to Hard power down the virtual machine,
    this is equivalent to pulling the power
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(side_effect=[False, ["a", "b", "c"]])
        with patch.object(xapi, "_get_label_uuid", mock):
            assert not xapi.stop("salt")
            assert xapi.stop("salt")

    with patch.object(xapi, "_check_xenapi", MagicMock(return_value=Mockxapi)):
        mock = MagicMock(return_value=True)
        with patch.dict(xapi.__salt__, {"config.option": mock}):
            with patch.object(xapi, "_get_label_uuid", mock):
                assert not xapi.stop("salt")


def test_is_hyper():
    """
    Test to returns a bool whether or not
    this node is a hypervisor of any kind
    """
    with patch.dict(xapi.__grains__, {"virtual_subtype": "Dom0"}):
        assert not xapi.is_hyper()
    with patch.dict(xapi.__grains__, {"virtual": "Xen Dom0"}):
        assert not xapi.is_hyper()
    with patch.dict(xapi.__grains__, {"virtual_subtype": "Xen Dom0"}):
        with patch("salt.utils.files.fopen", mock_open(read_data="salt")):
            assert not xapi.is_hyper()
        with patch("salt.utils.files.fopen", mock_open()) as mock_read:
            mock_read.side_effect = IOError
            assert not xapi.is_hyper()
        with patch("salt.utils.files.fopen", mock_open(read_data="xen_")):
            with patch.dict(xapi.__grains__, {"ps": "salt"}):
                mock = MagicMock(return_value={"xenstore": "salt"})
                with patch.dict(xapi.__salt__, {"cmd.run": mock}):
                    assert xapi.is_hyper()


def test_vm_cputime():
    """
    Test to Return cputime used by the vms
    """
    ret = {"1": {"cputime_percent": 0, "cputime": 1}}
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        mock = MagicMock(return_value={"host_CPUs": "1"})
        with patch.object(xapi, "_get_record_by_label", mock):
            mock = MagicMock(
                return_value={"VCPUs_number": "1", "VCPUs_utilisation": {"0": "1"}}
            )
            with patch.object(xapi, "_get_metrics_record", mock):
                assert xapi.vm_cputime("1") == ret
        mock = MagicMock(return_value={})
        with patch.object(xapi, "list_domains", mock):
            assert xapi.vm_cputime("") == {}


def test_vm_netstats():
    """
    Test to return combined network counters used by the vms
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        assert xapi.vm_netstats("") == {}


def test_vm_diskstats():
    """
    Test to return disk usage counters used by the vms
    """
    with patch.object(xapi, "_get_xapi_session", MagicMock()):
        assert xapi.vm_diskstats("") == {}



"""
unit tests for the cache runner
"""

import pytest

import salt.runners.queue as queue_mod
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(tmp_path):
    return {
        queue_mod: {
            "__opts__": {
                "sock_dir": str(tmp_path / "queue-runner-sock-dir"),
                "transport": "zeromq",
            }
        }
    }


def test_insert_runner():
    queue_insert = MagicMock(return_value=True)
    with patch.object(queue_mod, "insert", queue_insert):
        queue_mod.insert_runner("test.stdout_print", queue="salt")
    expected_call = {
        "queue": "salt",
        "items": {"fun": "test.stdout_print", "args": [], "kwargs": {}},
        "backend": "pgjsonb",
    }
    queue_insert.assert_called_once_with(**expected_call)


@pytest.mark.slow_test
def test_process_runner():
    ret = [{"fun": "test.stdout_print", "args": [], "kwargs": {}}]

    queue_pop = MagicMock(return_value=ret)
    test_stdout_print = MagicMock(return_value=True)
    with patch.dict(queue_mod.__salt__, {"test.stdout_print": test_stdout_print}):
        with patch.object(queue_mod, "pop", queue_pop):
            queue_mod.process_runner(queue="salt")
        queue_pop.assert_called_once_with(
            is_runner=True, queue="salt", quantity=1, backend="pgjsonb"
        )
        test_stdout_print.assert_called_once_with()
        queue_pop.assert_called_once_with(
            is_runner=True, queue="salt", quantity=1, backend="pgjsonb"
        )



"""
Create virtualenv environments.

.. versionadded:: 0.17.0
"""

import glob
import logging
import os
import re
import shutil
import sys

import salt.utils.files
import salt.utils.path
import salt.utils.platform
import salt.utils.verify
from salt.exceptions import CommandExecutionError, SaltInvocationError

KNOWN_BINARY_NAMES = frozenset(
    [
        "virtualenv-{}.{}".format(*sys.version_info[:2]),
        f"virtualenv{sys.version_info[0]}",
        "virtualenv",
    ]
)

log = logging.getLogger(__name__)

__opts__ = {"venv_bin": salt.utils.path.which_bin(KNOWN_BINARY_NAMES) or "venv"}

__pillar__ = {}

# Define the module's virtual name
__virtualname__ = "virtualenv"


def __virtual__():
    return __virtualname__


def virtualenv_ver(venv_bin, user=None, **kwargs):
    """
    return virtualenv version if exists
    """
    # Virtualenv package
    try:
        import virtualenv

        version = getattr(virtualenv, "__version__", None)
        if not version:
            version = virtualenv.virtualenv_version
    except ImportError:
        # Unable to import?? Let's parse the version from the console
        version_cmd = [venv_bin, "--version"]
        ret = __salt__["cmd.run_all"](
            version_cmd, runas=user, python_shell=False, redirect_stderr=True, **kwargs
        )
        if ret["retcode"] > 0 or not ret["stdout"].strip():
            raise CommandExecutionError(
                "Unable to get the virtualenv version output using '{}'. "
                "Returned data: {}".format(version_cmd, ret)
            )
        # 20.0.0 virtualenv changed the --version output. find version number
        version = "".join(
            [x for x in ret["stdout"].strip().split() if re.search(r"^\d.\d*", x)]
        )
    virtualenv_version_info = tuple(
        int(i) for i in re.sub(r"(rc|\+ds).*$", "", version).split(".")
    )
    return virtualenv_version_info


def create(
    path,
    venv_bin=None,
    system_site_packages=False,
    distribute=False,
    clear=False,
    python=None,
    extra_search_dir=None,
    never_download=None,
    prompt=None,
    pip=False,
    symlinks=None,
    upgrade=None,
    user=None,
    use_vt=False,
    saltenv="base",
    **kwargs,
):
    """
    Create a virtualenv

    path
        The path to the virtualenv to be created

    venv_bin
        The name (and optionally path) of the virtualenv command. This can also
        be set globally in the minion config file as ``virtualenv.venv_bin``.
        Defaults to ``virtualenv``.

    system_site_packages : False
        Passthrough argument given to virtualenv or venv

    distribute : False
        Passthrough argument given to virtualenv

    pip : False
        Install pip after creating a virtual environment. Implies
        ``distribute=True``

    clear : False
        Passthrough argument given to virtualenv or venv

    python : None (default)
        Passthrough argument given to virtualenv

    extra_search_dir : None (default)
        Passthrough argument given to virtualenv

    never_download : None (default)
        Passthrough argument given to virtualenv if True

    prompt : None (default)
        Passthrough argument given to virtualenv if not None

    symlinks : None
        Passthrough argument given to venv if True

    upgrade : None
        Passthrough argument given to venv if True

    user : None
        Set ownership for the virtualenv

        .. note::
            On Windows you must also pass a ``password`` parameter. Additionally,
            the user must have permissions to the location where the virtual
            environment is being created

    runas : None
        Set ownership for the virtualenv

        .. deprecated:: 2014.1.0
            ``user`` should be used instead

    use_vt : False
        Use VT terminal emulation (see output while installing)

        .. versionadded:: 2015.5.0

    saltenv : 'base'
        Specify a different environment. The default environment is ``base``.

        .. versionadded:: 2014.1.0

    .. note::
        The ``runas`` argument is deprecated as of 2014.1.0. ``user`` should be
        used instead.

    CLI Example:

    .. code-block:: console

        salt '*' virtualenv.create /path/to/new/virtualenv

     Example of using --always-copy environment variable (in case your fs doesn't support symlinks).
     This will copy files into the virtualenv instead of symlinking them.

     .. code-block:: yaml

         - env:
           - VIRTUALENV_ALWAYS_COPY: 1
    """
    if venv_bin is None:
        venv_bin = __pillar__.get("venv_bin") or __opts__.get("venv_bin")

    if venv_bin == "venv":
        cmd = [sys.executable, "-m", "venv"]
    else:
        cmd = [venv_bin]

    if "venv" not in venv_bin:
        # ----- Stop the user if venv only options are used ----------------->
        # If any of the following values are not None, it means that the user
        # is actually passing a True or False value. Stop Him!
        if upgrade is not None:
            raise CommandExecutionError(
                "The `upgrade`(`--upgrade`) option is not supported by '{}'".format(
                    venv_bin
                )
            )
        elif symlinks is not None:
            raise CommandExecutionError(
                "The `symlinks`(`--symlinks`) option is not supported by '{}'".format(
                    venv_bin
                )
            )
        # <---- Stop the user if venv only options are used ------------------

        virtualenv_version_info = virtualenv_ver(venv_bin, user=user, **kwargs)

        if distribute:
            if virtualenv_version_info >= (1, 10):
                log.info(
                    "The virtualenv '--distribute' option has been "
                    "deprecated in virtualenv(>=1.10), as such, the "
                    "'distribute' option to `virtualenv.create()` has "
                    "also been deprecated and it's not necessary anymore."
                )
            else:
                cmd.append("--distribute")

        if python is not None and python.strip() != "":
            if not salt.utils.path.which(python):
                raise CommandExecutionError(f"Cannot find requested python ({python}).")
            cmd.append(f"--python={python}")
        if extra_search_dir is not None:
            if isinstance(extra_search_dir, str) and extra_search_dir.strip() != "":
                extra_search_dir = [e.strip() for e in extra_search_dir.split(",")]
            for entry in extra_search_dir:
                cmd.append(f"--extra-search-dir={entry}")
        if never_download is True:
            if (1, 10) <= virtualenv_version_info < (14, 0, 0):
                log.info(
                    "--never-download was deprecated in 1.10.0, but reimplemented in"
                    " 14.0.0. If this feature is needed, please install a supported"
                    " virtualenv version."
                )
            else:
                cmd.append("--never-download")
        if prompt is not None and prompt.strip() != "":
            cmd.append(f"--prompt='{prompt}'")
    else:
        # venv module from the Python >= 3.3 standard library

        # ----- Stop the user if virtualenv only options are being used ----->
        # If any of the following values are not None, it means that the user
        # is actually passing a True or False value. Stop Him!
        if python is not None and python.strip() != "":
            raise CommandExecutionError(
                "The `python`(`--python`) option is not supported by '{}'".format(
                    venv_bin
                )
            )
        elif extra_search_dir is not None and extra_search_dir.strip() != "":
            raise CommandExecutionError(
                "The `extra_search_dir`(`--extra-search-dir`) option is not "
                "supported by '{}'".format(venv_bin)
            )
        elif never_download is not None:
            raise CommandExecutionError(
                "The `never_download`(`--never-download`) option is not "
                "supported by '{}'".format(venv_bin)
            )
        elif prompt is not None and prompt.strip() != "":
            raise CommandExecutionError(
                "The `prompt`(`--prompt`) option is not supported by '{}'".format(
                    venv_bin
                )
            )
        # <---- Stop the user if virtualenv only options are being used ------

        if upgrade is True:
            cmd.append("--upgrade")
        if symlinks is True:
            cmd.append("--symlinks")

    # Common options to virtualenv and venv
    if clear is True:
        cmd.append("--clear")
    if system_site_packages is True:
        cmd.append("--system-site-packages")

    # Finally the virtualenv path
    cmd.append(path)

    # Let's create the virtualenv
    ret = __salt__["cmd.run_all"](cmd, runas=user, python_shell=False, **kwargs)
    if ret["retcode"] != 0:
        # Something went wrong. Let's bail out now!
        return ret

    # Check if distribute and pip are already installed
    if salt.utils.platform.is_windows():
        venv_python = os.path.join(path, "Scripts", "python.exe")
        venv_pip = os.path.join(path, "Scripts", "pip.exe")
        venv_setuptools = os.path.join(path, "Scripts", "easy_install.exe")
    else:
        venv_python = os.path.join(path, "bin", "python")
        venv_pip = os.path.join(path, "bin", "pip")
        venv_setuptools = os.path.join(path, "bin", "easy_install")

    # Install setuptools
    if (pip or distribute) and not os.path.exists(venv_setuptools):
        _install_script(
            "https://bootstrap.pypa.io/ez_setup.py",
            path,
            venv_python,
            user,
            saltenv=saltenv,
            use_vt=use_vt,
        )

        # clear up the distribute archive which gets downloaded
        for fpath in glob.glob(os.path.join(path, "distribute-*.tar.gz*")):
            os.unlink(fpath)

    if ret["retcode"] != 0:
        # Something went wrong. Let's bail out now!
        return ret

    # Install pip
    if pip and not os.path.exists(venv_pip):
        _ret = _install_script(
            "https://bootstrap.pypa.io/get-pip.py",
            path,
            venv_python,
            user,
            saltenv=saltenv,
            use_vt=use_vt,
        )
        # Let's update the return dictionary with the details from the pip
        # installation
        ret.update(
            retcode=_ret["retcode"],
            stdout="{}\n{}".format(ret["stdout"], _ret["stdout"]).strip(),
            stderr="{}\n{}".format(ret["stderr"], _ret["stderr"]).strip(),
        )

    return ret


def get_site_packages(venv):
    """
    Return the path to the site-packages directory of a virtualenv

    venv
        Path to the virtualenv.

    CLI Example:

    .. code-block:: bash

        salt '*' virtualenv.get_site_packages /path/to/my/venv
    """
    bin_path = _verify_virtualenv(venv)

    # note: platlib and purelib could differ
    ret = __salt__["cmd.exec_code_all"](
        bin_path,
        'import sysconfig; print(sysconfig.get_path("purelib"))',
    )

    if ret["retcode"] != 0:
        raise CommandExecutionError("{stdout}\n{stderr}".format(**ret))

    return ret["stdout"]


def get_distribution_path(venv, distribution):
    """
    Return the path to a distribution installed inside a virtualenv

    .. versionadded:: 2016.3.0

    venv
        Path to the virtualenv.
    distribution
        Name of the distribution. Note, all non-alphanumeric characters
        will be converted to dashes.

    CLI Example:

    .. code-block:: bash

        salt '*' virtualenv.get_distribution_path /path/to/my/venv my_distribution
    """
    _verify_safe_py_code(distribution)
    bin_path = _verify_virtualenv(venv)

    ret = __salt__["cmd.exec_code_all"](
        bin_path,
        "import pkg_resources; "
        "print(pkg_resources.get_distribution('{}').location)".format(distribution),
    )

    if ret["retcode"] != 0:
        raise CommandExecutionError("{stdout}\n{stderr}".format(**ret))

    return ret["stdout"]


def get_resource_path(venv, package=None, resource=None):
    """
    Return the path to a package resource installed inside a virtualenv

    .. versionadded:: 2015.5.0

    venv
        Path to the virtualenv

    package
        Name of the package in which the resource resides

        .. versionadded:: 2016.3.0

    resource
        Name of the resource of which the path is to be returned

        .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' virtualenv.get_resource_path /path/to/my/venv my_package my/resource.xml
    """
    _verify_safe_py_code(package, resource)
    bin_path = _verify_virtualenv(venv)

    ret = __salt__["cmd.exec_code_all"](
        bin_path,
        "import pkg_resources; "
        "print(pkg_resources.resource_filename('{}', '{}'))".format(package, resource),
    )

    if ret["retcode"] != 0:
        raise CommandExecutionError("{stdout}\n{stderr}".format(**ret))

    return ret["stdout"]


def get_resource_content(venv, package=None, resource=None):
    """
    Return the content of a package resource installed inside a virtualenv

    .. versionadded:: 2015.5.0

    venv
        Path to the virtualenv

    package
        Name of the package in which the resource resides

        .. versionadded:: 2016.3.0

    resource
        Name of the resource of which the content is to be returned

        .. versionadded:: 2016.3.0

    CLI Example:

    .. code-block:: bash

        salt '*' virtualenv.get_resource_content /path/to/my/venv my_package my/resource.xml
    """
    _verify_safe_py_code(package, resource)
    bin_path = _verify_virtualenv(venv)

    ret = __salt__["cmd.exec_code_all"](
        bin_path,
        "import pkg_resources; print(pkg_resources.resource_string('{}', '{}'))".format(
            package, resource
        ),
    )

    if ret["retcode"] != 0:
        raise CommandExecutionError("{stdout}\n{stderr}".format(**ret))

    return ret["stdout"]


def _install_script(source, cwd, python, user, saltenv="base", use_vt=False):
    if not salt.utils.platform.is_windows():
        tmppath = salt.utils.files.mkstemp(dir=cwd)
    else:
        tmppath = __salt__["cp.cache_file"](source, saltenv)

    if not salt.utils.platform.is_windows():
        fn_ = __salt__["cp.cache_file"](source, saltenv)
        shutil.copyfile(fn_, tmppath)
        os.chmod(tmppath, 0o500)
        os.chown(tmppath, __salt__["file.user_to_uid"](user), -1)
    try:
        return __salt__["cmd.run_all"](
            [python, tmppath],
            runas=user,
            cwd=cwd,
            env={"VIRTUAL_ENV": cwd},
            use_vt=use_vt,
            python_shell=False,
        )
    finally:
        os.remove(tmppath)


def _verify_safe_py_code(*args):
    for arg in args:
        if not salt.utils.verify.safe_py_code(arg):
            raise SaltInvocationError(f"Unsafe python code detected in '{arg}'")


def _verify_virtualenv(venv_path):
    if salt.utils.platform.is_windows():
        bin_path = os.path.join(venv_path, "Scripts", "python.exe")
    else:
        bin_path = os.path.join(venv_path, "bin", "python")

    if not os.path.exists(bin_path):
        raise CommandExecutionError(
            f"Path '{venv_path}' does not appear to be a virtualenv: '{bin_path}' not found."
        )
    return bin_path



import os

import pytest

import salt.utils.boto3mod as boto3mod
import salt.utils.botomod as botomod
from salt.exceptions import SaltInvocationError
from salt.utils.versions import Version
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import TestCase

# pylint: disable=import-error
try:
    import boto

    boto.ENDPOINTS_PATH = os.path.join(
        RUNTIME_VARS.TESTS_DIR, "unit/files/endpoints.json"
    )
    import boto.exception
    from boto.exception import BotoServerError

    HAS_BOTO = True
except ImportError:
    HAS_BOTO = False

try:
    import boto3

    HAS_BOTO3 = True
except ImportError:
    HAS_BOTO3 = False

try:
    from moto import mock_ec2

    HAS_MOTO = True
except ImportError:
    HAS_MOTO = False

    def mock_ec2(self):
        """
        if the mock_ec2 function is not available due to import failure
        this replaces the decorated function with stub_function.
        Allows unit tests to use the @mock_ec2 decorator
        without a "NameError: name 'mock_ec2' is not defined" error.
        """

        def stub_function(self):
            pass

        return stub_function


pytestmark = [
    pytest.mark.skip_on_fips_enabled_platform,
]


required_boto_version = "2.0.0"
required_boto3_version = "1.2.1"
region = "us-east-1"
access_key = "GKTADJGHEIQSXMKKRBJ08H"
secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
conn_parameters = {
    "region": region,
    "key": access_key,
    "keyid": secret_key,
    "profile": {},
}

service = "ec2"
resource_name = "test-instance"
resource_id = "i-a1b2c3"


error_body = """
<Response>
    <Errors>
         <Error>
           <Code>Error code text</Code>
           <Message>Error message</Message>
         </Error>
    </Errors>
    <RequestID>request ID</RequestID>
</Response>
"""

no_error_body = """
<Response>
    <Errors />
    <RequestID>request ID</RequestID>
</Response>
"""


def _has_required_boto():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    if not HAS_BOTO:
        return False
    elif Version(boto.__version__) < Version(required_boto_version):
        return False
    else:
        return True


def _has_required_boto3():
    """
    Returns True/False boolean depending on if Boto is installed and correct
    version.
    """
    try:
        if not HAS_BOTO3:
            return False
        elif Version(boto3.__version__) < Version(required_boto3_version):
            return False
        else:
            return True
    except AttributeError as exc:
        if "has no attribute '__version__'" not in str(exc):
            raise
        return False


def _has_required_moto():
    """
    Returns True/False boolean depending on if Moto is installed and correct
    version.
    """
    if not HAS_MOTO:
        return False
    else:
        import pkg_resources

        if Version(pkg_resources.get_distribution("moto").version) < Version("0.3.7"):
            return False
        return True


class BotoUtilsTestCaseBase(TestCase, LoaderModuleMockMixin):
    def setup_loader_modules(self):
        module_globals = {
            "__salt__": {"config.option": MagicMock(return_value="dummy_opt")}
        }
        return {botomod: module_globals, boto3mod: module_globals}


class BotoUtilsCacheIdTestCase(BotoUtilsTestCaseBase):
    def test_set_and_get_with_no_auth_params(self):
        botomod.cache_id(service, resource_name, resource_id=resource_id)
        self.assertEqual(botomod.cache_id(service, resource_name), resource_id)

    def test_set_and_get_with_explicit_auth_params(self):
        botomod.cache_id(
            service, resource_name, resource_id=resource_id, **conn_parameters
        )
        self.assertEqual(
            botomod.cache_id(service, resource_name, **conn_parameters), resource_id
        )

    def test_set_and_get_with_different_region_returns_none(self):
        botomod.cache_id(
            service, resource_name, resource_id=resource_id, region="us-east-1"
        )
        self.assertEqual(
            botomod.cache_id(service, resource_name, region="us-west-2"), None
        )

    def test_set_and_get_after_invalidation_returns_none(self):
        botomod.cache_id(service, resource_name, resource_id=resource_id)
        botomod.cache_id(
            service, resource_name, resource_id=resource_id, invalidate=True
        )
        self.assertEqual(botomod.cache_id(service, resource_name), None)

    def test_partial(self):
        cache_id = botomod.cache_id_func(service)
        cache_id(resource_name, resource_id=resource_id)
        self.assertEqual(cache_id(resource_name), resource_id)


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(HAS_MOTO is False, reason="The moto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}".format(
        required_boto_version
    ),
)
class BotoUtilsGetConnTestCase(BotoUtilsTestCaseBase):
    @mock_ec2
    def test_conn_is_cached(self):
        conn = botomod.get_connection(service, **conn_parameters)
        self.assertTrue(conn in botomod.__context__.values())

    @mock_ec2
    def test_conn_is_cache_with_profile(self):
        conn = botomod.get_connection(service, profile=conn_parameters)
        self.assertTrue(conn in botomod.__context__.values())

    @mock_ec2
    def test_get_conn_with_no_auth_params_raises_invocation_error(self):
        with patch(
            f"boto.{service}.connect_to_region",
            side_effect=boto.exception.NoAuthHandlerFound(),
        ):
            with self.assertRaises(SaltInvocationError):
                botomod.get_connection(service)

    @mock_ec2
    def test_get_conn_error_raises_command_execution_error(self):
        with patch(
            f"boto.{service}.connect_to_region",
            side_effect=BotoServerError(400, "Mocked error", body=error_body),
        ):
            with self.assertRaises(BotoServerError):
                botomod.get_connection(service)

    @mock_ec2
    def test_partial(self):
        get_conn = botomod.get_connection_func(service)
        conn = get_conn(**conn_parameters)
        self.assertTrue(conn in botomod.__context__.values())


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}".format(
        required_boto_version
    ),
)
class BotoUtilsGetErrorTestCase(BotoUtilsTestCaseBase):
    def test_error_message(self):
        e = BotoServerError("400", "Mocked error", body=error_body)
        r = botomod.get_error(e)
        expected = {
            "aws": {
                "code": "Error code text",
                "message": "Error message",
                "reason": "Mocked error",
                "status": "400",
            },
            "message": "Mocked error: Error message",
        }
        self.assertEqual(r, expected)

    def test_exception_message_with_no_body(self):
        e = BotoServerError("400", "Mocked error")
        r = botomod.get_error(e)
        expected = {
            "aws": {"reason": "Mocked error", "status": "400"},
            "message": "Mocked error",
        }
        self.assertEqual(r, expected)

    def test_exception_message_with_no_error_in_body(self):
        e = BotoServerError("400", "Mocked error", body=no_error_body)
        r = botomod.get_error(e)
        expected = {
            "aws": {"reason": "Mocked error", "status": "400"},
            "message": "Mocked error",
        }
        self.assertEqual(r, expected)


@pytest.mark.skipif(HAS_BOTO is False, reason="The boto module must be installed.")
@pytest.mark.skipif(
    _has_required_boto() is False,
    reason="The boto module must be greater than or equal to version {}".format(
        required_boto_version
    ),
)
@pytest.mark.skipif(HAS_BOTO3 is False, reason="The boto3 module must be installed.")
@pytest.mark.skipif(
    _has_required_boto3() is False,
    reason="The boto3 module must be greater than or equal to version {}".format(
        required_boto3_version
    ),
)
class BotoBoto3CacheContextCollisionTest(BotoUtilsTestCaseBase):
    def test_context_conflict_between_boto_and_boto3_utils(self):
        botomod.assign_funcs(__name__, "ec2")
        boto3mod.assign_funcs(__name__, "ec2", get_conn_funcname="_get_conn3")

        boto_ec2_conn = botomod.get_connection(
            "ec2", region=region, key=secret_key, keyid=access_key
        )
        boto3_ec2_conn = boto3mod.get_connection(
            "ec2", region=region, key=secret_key, keyid=access_key
        )

        # These should *not* be the same object!
        self.assertNotEqual(id(boto_ec2_conn), id(boto3_ec2_conn))



import collections
import logging

import pytest

import salt.serializers.json as jsonserializer
import salt.serializers.msgpack as msgpackserializer
import salt.serializers.plist as plistserializer
import salt.serializers.python as pythonserializer
import salt.serializers.yaml as yamlserializer
import salt.states.file as filestate
from tests.support.helpers import dedent

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        filestate: {
            "__env__": "base",
            "__salt__": {"file.manage_file": False},
            "__serializers__": {
                "yaml.serialize": yamlserializer.serialize,
                "yaml.seserialize": yamlserializer.serialize,
                "python.serialize": pythonserializer.serialize,
                "json.serialize": jsonserializer.serialize,
                "plist.serialize": plistserializer.serialize,
                "msgpack.serialize": msgpackserializer.serialize,
            },
            "__opts__": {"test": False, "cachedir": ""},
            "__instance_id__": "",
            "__low__": {},
            "__utils__": {},
        }
    }


def test_file_keyvalue_key_values(tmp_path):
    """
    test file.keyvalue when using key_values kwarg
    """
    contents = dedent(
        """\
        #PermitRootLogin prohibit-password
        #StrictMode yes
        """
    )
    with pytest.helpers.temp_file(
        "tempfile", contents=contents, directory=tmp_path
    ) as tempfile:
        ret = filestate.keyvalue(
            name=str(tempfile),
            key_values=collections.OrderedDict(PermitRootLogin="yes"),
            separator=" ",
            uncomment="#",
            key_ignore_case=True,
        )

        f_contents = tempfile.read_text()
        assert "PermitRootLogin yes" in f_contents
        assert "#StrictMode yes" in f_contents


def test_file_keyvalue_empty(tmp_path):
    """
    test file.keyvalue when key_values is empty
    """
    contents = dedent(
        """\
        #PermitRootLogin prohibit-password
        #StrictMode yes
        """
    )
    with pytest.helpers.temp_file(
        "tempfile", contents=contents, directory=tmp_path
    ) as tempfile:
        ret = filestate.keyvalue(
            name=str(tempfile),
            key_values={},
            separator=" ",
            uncomment="#",
            key_ignore_case=True,
        )

        assert (
            ret["comment"]
            == "file.keyvalue key and value not supplied and key_values is empty"
        )
        f_contents = tempfile.read_text()
        assert "PermitRootLogin yes" not in f_contents
        assert "#StrictMode yes" in f_contents


def test_file_keyvalue_not_dict(tmp_path):
    """
    test file.keyvalue when key_values not a dict
    """
    contents = dedent(
        """\
        #PermitRootLogin prohibit-password
        #StrictMode yes
        """
    )
    with pytest.helpers.temp_file(
        "tempfile", contents=contents, directory=tmp_path
    ) as tempfile:
        ret = filestate.keyvalue(
            name=str(tempfile),
            key_values=["PermiteRootLogin", "yes"],
            separator=" ",
            uncomment="#",
            key_ignore_case=True,
        )

        assert (
            ret["comment"]
            == "file.keyvalue key and value not supplied and key_values is not a dictionary"
        )
        f_contents = tempfile.read_text()
        assert "PermitRootLogin yes" not in f_contents
        assert "#StrictMode yes" in f_contents



"""
Venafi Pillar Certificates

This module will only return pillar data if the ``venafi`` runner module has
already been used to create certificates.

To configure this module, set ``venafi`` to ``True`` in the ``ext_pillar``
section of your ``master`` configuration file:

.. code-block:: yaml

    ext_pillar:
      - venafi: True
"""

import logging

import salt.cache
import salt.syspaths as syspaths

__virtualname__ = "venafi"
log = logging.getLogger(__name__)


def __virtual__():
    """
    No special requirements outside of Salt itself
    """
    return __virtualname__


def ext_pillar(minion_id, pillar, conf):
    """
    Return an existing set of certificates
    """
    cache = salt.cache.Cache(__opts__, syspaths.CACHE_DIR)

    ret = {}
    for dns_name in cache.list("venafi/domains"):
        data = cache.fetch("venafi/domains", dns_name)
        if data["minion_id"] == minion_id:
            ret[dns_name] = data
    return {"venafi": ret}



import os
import types

import pytest

import salt.spm
import salt.utils.files
from tests.support.mock import patch


@pytest.fixture
def formula_definition():
    return {
        "name": "formula1",
        "version": "1.2",
        "release": "2",
        "summary": "test",
        "description": "testing, nothing to see here",
    }


@pytest.fixture
def formula_contents(formula_definition):
    return (
        (
            "FORMULA",
            (
                "name: {name}\n"
                "version: {version}\n"
                "release: {release}\n"
                "summary: {summary}\n"
                "description: {description}"
            ).format(**formula_definition),
        ),
        ("modules/mod1.py", "# mod1.py"),
        ("modules/mod2.py", "# mod2.py"),
        ("states/state1.sls", "# state1.sls"),
        ("states/state2.sls", "# state2.sls"),
    )


@pytest.fixture
def formula(formula_definition, formula_contents):
    return types.SimpleNamespace(
        definition=formula_definition, contents=formula_contents
    )


class SPMTestUserInterface(salt.spm.SPMUserInterface):
    """
    Unit test user interface to SPMClient
    """

    def __init__(self):
        self._status = []
        self._confirm = []
        self._error = []

    def status(self, msg):
        self._status.append(msg)

    def confirm(self, action):
        self._confirm.append(action)

    def error(self, msg):
        self._error.append(msg)


@pytest.fixture
def minion_config(tmp_path, minion_opts):
    _minion_config = minion_opts.copy()
    _minion_config.update(
        {
            "spm_logfile": str(tmp_path / "log"),
            "spm_repos_config": str(tmp_path / "etc" / "spm.repos"),
            "spm_cache_dir": str(tmp_path / "cache"),
            "spm_build_dir": str(tmp_path / "build"),
            "spm_build_exclude": [".git"],
            "spm_db_provider": "sqlite3",
            "spm_files_provider": "local",
            "spm_db": str(tmp_path / "packages.db"),
            "extension_modules": str(tmp_path / "modules"),
            "file_roots": {"base": [str(tmp_path)]},
            "formula_path": str(tmp_path / "spm"),
            "pillar_path": str(tmp_path / "pillar"),
            "reactor_path": str(tmp_path / "reactor"),
            "assume_yes": True,
            "root_dir": str(tmp_path),
            "force": False,
            "verbose": False,
            "cache": "localfs",
            "cachedir": str(tmp_path / "cache"),
            "spm_repo_dups": "ignore",
            "spm_share_dir": str(tmp_path / "share"),
        }
    )
    return _minion_config


@pytest.fixture
def client(minion_config):
    with patch("salt.client.Caller", return_value=minion_config):
        with patch(
            "salt.client.get_local_client", return_value=minion_config["conf_file"]
        ):
            yield salt.spm.SPMClient(SPMTestUserInterface(), minion_config)


@pytest.fixture
def formulas_dir(formula, tmp_path):
    fdir = tmp_path / formula.definition["name"]
    fdir.mkdir()
    for path, contents in formula.contents:
        path = fdir / path
        dirname, _ = os.path.split(str(path))
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        path.write_text(contents)
    return str(fdir)


def test_build_install(client, formulas_dir, minion_config, formula):
    # Build package
    client.run(["build", formulas_dir])
    pkgpath = client.ui._status[-1].split()[-1]
    assert os.path.exists(pkgpath)
    # Install package
    client.run(["local", "install", pkgpath])
    # Check filesystem
    for path, contents in formula.contents:
        path = os.path.join(
            minion_config["file_roots"]["base"][0],
            formula.definition["name"],
            path,
        )
        assert os.path.exists(path)
        with salt.utils.files.fopen(path, "r") as rfh:
            assert rfh.read().replace("\r\n", "\n") == contents
    # Check database
    client.run(["info", formula.definition["name"]])
    lines = client.ui._status[-1].replace("\r\n", "\n").split("\n")
    for key, line in (
        ("name", "Name: {}"),
        ("version", "Version: {}"),
        ("release", "Release: {}"),
        ("summary", "Summary: {}"),
    ):
        assert line.format(formula.definition[key]) in lines
    # Reinstall with force=False, should fail
    client.ui._error = []
    client.run(["local", "install", pkgpath])
    assert len(client.ui._error) > 0
    # Reinstall with force=True, should succeed
    with patch.dict(minion_config, {"force": True}):
        client.ui._error = []
        client.run(["local", "install", pkgpath])
        assert len(client.ui._error) == 0


def test_repo_paths(client, formulas_dir):
    client.run(["create_repo", formulas_dir])
    assert len(client.ui._error) == 0


@pytest.mark.parametrize(
    "fail_args",
    (
        ["bogus", "command"],
        ["create_repo"],
        ["build"],
        ["build", "/nonexistent/path"],
        ["info"],
        ["info", "not_installed"],
        ["files"],
        ["files", "not_installed"],
        ["install"],
        ["install", "nonexistent.spm"],
        ["remove"],
        ["remove", "not_installed"],
        ["local", "bogus", "command"],
        ["local", "info"],
        ["local", "info", "/nonexistent/path/junk.spm"],
        ["local", "files"],
        ["local", "files", "/nonexistent/path/junk.spm"],
        ["local", "install"],
        ["local", "install", "/nonexistent/path/junk.spm"],
        ["local", "list"],
        ["local", "list", "/nonexistent/path/junk.spm"],
        # XXX install failure due to missing deps
        # XXX install failure due to missing field
    ),
)
def test_failure_paths(client, fail_args):
    client.run(fail_args)
    assert len(client.ui._error) > 0



import copy
import logging
import os
import random

import salt.client.ssh
import salt.config
import salt.syspaths
import salt.utils.args
from salt.exceptions import SaltClientError

log = logging.getLogger(__name__)


class SSHClient:
    """
    Create a client object for executing routines via the salt-ssh backend

    .. versionadded:: 2015.5.0
    """

    def __init__(
        self,
        c_path=os.path.join(salt.syspaths.CONFIG_DIR, "master"),
        mopts=None,
        disable_custom_roster=False,
    ):
        if mopts:
            self.opts = mopts
        else:
            if os.path.isdir(c_path):
                log.warning(
                    "%s expects a file path not a directory path(%s) to "
                    "its 'c_path' keyword argument",
                    self.__class__.__name__,
                    c_path,
                )
            self.opts = salt.config.client_config(c_path)

        # Salt API should never offer a custom roster!
        self.opts["__disable_custom_roster"] = disable_custom_roster
        # Pillar compilation and nested SSH calls require the correct config_dir
        # in __opts__, otherwise we will use the SSH minion's one later.
        if "config_dir" not in self.opts:
            self.opts["config_dir"] = os.path.dirname(c_path)

    def sanitize_kwargs(self, kwargs):
        roster_vals = [
            ("host", str),
            ("ssh_user", str),
            ("ssh_passwd", str),
            ("ssh_port", int),
            ("ssh_sudo", bool),
            ("ssh_sudo_user", str),
            ("ssh_priv", str),
            ("ssh_priv_passwd", str),
            ("ssh_identities_only", bool),
            ("ssh_remote_port_forwards", str),
            ("ssh_options", list),
            ("ssh_max_procs", int),
            ("ssh_askpass", bool),
            ("ssh_key_deploy", bool),
            ("ssh_update_roster", bool),
            ("ssh_scan_ports", str),
            ("ssh_scan_timeout", int),
            ("ssh_timeout", int),
            ("ssh_log_file", str),
            ("raw_shell", bool),
            ("refresh_cache", bool),
            ("roster", str),
            ("roster_file", str),
            ("rosters", list),
            ("ignore_host_keys", bool),
            ("raw_shell", bool),
            ("extra_filerefs", str),
            ("min_extra_mods", str),
            ("thin_extra_mods", str),
            ("verbose", bool),
            ("static", bool),
            ("ssh_wipe", bool),
            ("rand_thin_dir", bool),
            ("regen_thin", bool),
            ("ssh_run_pre_flight", bool),
            ("no_host_keys", bool),
            ("saltfile", str),
        ]
        sane_kwargs = {}
        for name, kind in roster_vals:
            if name not in kwargs:
                continue
            try:
                val = kind(kwargs[name])
            except ValueError:
                log.warning("Unable to cast kwarg %s", name)
                continue
            if kind is bool or kind is int:
                sane_kwargs[name] = val
            elif kind is str:
                if val.find("ProxyCommand") != -1:
                    log.warning("Filter unsafe value for kwarg %s", name)
                    continue
                sane_kwargs[name] = val
            elif kind is list:
                sane_val = []
                for item in val:
                    # This assumes the values are strings
                    if item.find("ProxyCommand") != -1:
                        log.warning("Filter unsafe value for kwarg %s", name)
                        continue
                    sane_val.append(item)
                sane_kwargs[name] = sane_val
        return sane_kwargs

    def _prep_ssh(
        self, tgt, fun, arg=(), timeout=None, tgt_type="glob", kwarg=None, **kwargs
    ):
        """
        Prepare the arguments
        """
        kwargs = self.sanitize_kwargs(kwargs)
        opts = copy.deepcopy(self.opts)
        opts.update(kwargs)
        if timeout:
            opts["timeout"] = timeout
        arg = salt.utils.args.condition_input(arg, kwarg)
        opts["argv"] = [fun] + arg
        opts["selected_target_option"] = tgt_type
        opts["tgt"] = tgt
        opts["arg"] = arg
        return salt.client.ssh.SSH(opts)

    def cmd_iter(
        self,
        tgt,
        fun,
        arg=(),
        timeout=None,
        tgt_type="glob",
        ret="",
        kwarg=None,
        **kwargs
    ):
        """
        Execute a single command via the salt-ssh subsystem and return a
        generator

        .. versionadded:: 2015.5.0
        """
        ssh = self._prep_ssh(tgt, fun, arg, timeout, tgt_type, kwarg, **kwargs)
        yield from ssh.run_iter(jid=kwargs.get("jid", None))

    def cmd(
        self, tgt, fun, arg=(), timeout=None, tgt_type="glob", kwarg=None, **kwargs
    ):
        """
        Execute a single command via the salt-ssh subsystem and return all
        routines at once

        .. versionadded:: 2015.5.0
        """
        ssh = self._prep_ssh(tgt, fun, arg, timeout, tgt_type, kwarg, **kwargs)
        final = {}
        for ret in ssh.run_iter(jid=kwargs.get("jid", None)):
            final.update(ret)
        return final

    def cmd_sync(self, low):
        """
        Execute a salt-ssh call synchronously.

        .. versionadded:: 2015.5.0

        WARNING: Eauth is **NOT** respected

        .. code-block:: python

            client.cmd_sync({
                'tgt': 'silver',
                'fun': 'test.ping',
                'arg': (),
                'tgt_type'='glob',
                'kwarg'={}
                })
            {'silver': {'fun_args': [], 'jid': '20141202152721523072', 'return': True, 'retcode': 0, 'success': True, 'fun': 'test.ping', 'id': 'silver'}}
        """

        kwargs = copy.deepcopy(low)

        for ignore in ["tgt", "fun", "arg", "timeout", "tgt_type", "kwarg"]:
            if ignore in kwargs:
                del kwargs[ignore]

        return self.cmd(
            low["tgt"],
            low["fun"],
            low.get("arg", []),
            low.get("timeout"),
            low.get("tgt_type"),
            low.get("kwarg"),
            **kwargs
        )

    def cmd_async(self, low, timeout=None):
        """
        Execute aa salt-ssh asynchronously

        WARNING: Eauth is **NOT** respected

        .. code-block:: python

            client.cmd_sync({
                'tgt': 'silver',
                'fun': 'test.ping',
                'arg': (),
                'tgt_type'='glob',
                'kwarg'={}
                })
            {'silver': {'fun_args': [], 'jid': '20141202152721523072', 'return': True, 'retcode': 0, 'success': True, 'fun': 'test.ping', 'id': 'silver'}}
        """
        # TODO Not implemented
        raise SaltClientError

    def cmd_subset(
        self,
        tgt,
        fun,
        arg=(),
        timeout=None,
        tgt_type="glob",
        ret="",
        kwarg=None,
        subset=3,
        **kwargs
    ):
        """
        Execute a command on a random subset of the targeted systems

        The function signature is the same as :py:meth:`cmd` with the
        following exceptions.

        :param subset: The number of systems to execute on

        .. code-block:: python

            >>> import salt.client.ssh.client
            >>> sshclient= salt.client.ssh.client.SSHClient()
            >>> sshclient.cmd_subset('*', 'test.ping', subset=1)
            {'jerry': True}

        .. versionadded:: 2017.7.0
        """
        minion_ret = self.cmd(tgt, "sys.list_functions", tgt_type=tgt_type, **kwargs)
        minions = list(minion_ret)
        random.shuffle(minions)
        f_tgt = []
        for minion in minions:
            if fun in minion_ret[minion]["return"]:
                f_tgt.append(minion)
            if len(f_tgt) >= subset:
                break
        return self.cmd_iter(
            f_tgt, fun, arg, timeout, tgt_type="list", ret=ret, kwarg=kwarg, **kwargs
        )

    def destroy(self):
        """
        API compatibility method with salt.client.LocalClient
        """

    def __enter__(self):
        """
        API compatibility method with salt.client.LocalClient
        """
        return self

    def __exit__(self, *args):
        """
        API compatibility method with salt.client.LocalClient
        """
        self.destroy()



"""
unit tests for the cache runner
"""

import pytest

import salt.runners.cache as cache
import salt.utils.master
from tests.support.mock import patch


@pytest.fixture
def configure_loader_modules(tmp_path):
    return {
        cache: {
            "__opts__": {
                "cache": "localfs",
                "pki_dir": str(tmp_path),
                "key_cache": True,
            }
        }
    }


def test_grains():
    """
    test cache.grains runner
    """
    mock_minion = ["Larry"]
    mock_ret = {}
    assert cache.grains(tgt="*", minion=mock_minion) == mock_ret

    mock_data = "grain stuff"

    class MockMaster:
        def __init__(self, *args, **kwargs):
            pass

        def get_minion_grains(self):
            return mock_data

    with patch.object(salt.utils.master, "MasterPillarUtil", MockMaster):
        assert cache.grains(tgt="*") == mock_data



import logging
import os

import pytest

import salt.modules.file as filemod
import salt.states.file as filestate
import salt.utils.files
import salt.utils.json
import salt.utils.platform
import salt.utils.win_functions
import salt.utils.yaml

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {filestate: {"__salt__": {"file.stats": filemod.stats}}}


@pytest.mark.destructive_test
@pytest.mark.skip_on_windows(reason="File modes do not exist on windows")
def test__check_directory(tmp_path):
    """
    Test the _check_directory function
    Make sure that recursive file permission checks return correctly
    """
    # set file permissions
    # Run _check_directory function
    # Verify that it returns correctly
    # Delete tmp directory structure
    root_tmp_dir = str(tmp_path / "test__check_dir")
    expected_mode = 0o770
    changed_mode = 0o755
    depth = 3

    def create_files(tmp_dir):
        for f in range(depth):
            path = os.path.join(tmp_dir, f"file_{f:03}.txt")
            with salt.utils.files.fopen(path, "w+"):
                os.chmod(path, expected_mode)

    # Create tmp directory structure
    os.mkdir(root_tmp_dir)
    os.chmod(root_tmp_dir, expected_mode)
    create_files(root_tmp_dir)

    for d in range(depth):
        dir_name = os.path.join(root_tmp_dir, f"dir{d:03}")
        os.mkdir(dir_name)
        os.chmod(dir_name, expected_mode)
        create_files(dir_name)
        for s in range(depth):
            sub_dir_name = os.path.join(dir_name, f"dir{s:03}")
            os.mkdir(sub_dir_name)
            os.chmod(sub_dir_name, expected_mode)
            create_files(sub_dir_name)
    # Symlinks on linux systems always have 0o777 permissions.
    # Ensure we are not treating them as modified files.
    target_dir = os.path.join(root_tmp_dir, "link_target_dir")
    target_file = os.path.join(target_dir, "link_target_file")
    link_dir = os.path.join(root_tmp_dir, "link_dir")
    link_to_dir = os.path.join(link_dir, "link_to_dir")
    link_to_file = os.path.join(link_dir, "link_to_file")

    os.mkdir(target_dir)
    os.mkdir(link_dir)
    with salt.utils.files.fopen(target_file, "w+"):
        pass
    os.symlink(target_dir, link_to_dir)
    os.symlink(target_file, link_to_file)
    for path in (target_dir, target_file, link_dir, link_to_dir, link_to_file):
        try:
            os.chmod(path, expected_mode, follow_symlinks=False)
        except (NotImplementedError, SystemError, OSError):
            os.chmod(path, expected_mode)

    # Set some bad permissions
    changed_files = {
        os.path.join(root_tmp_dir, "file_000.txt"),
        os.path.join(root_tmp_dir, "dir002", "file_000.txt"),
        os.path.join(root_tmp_dir, "dir000", "dir001", "file_002.txt"),
        os.path.join(root_tmp_dir, "dir001", "dir002"),
        os.path.join(root_tmp_dir, "dir002", "dir000"),
        os.path.join(root_tmp_dir, "dir001"),
    }
    for c in changed_files:
        os.chmod(c, changed_mode)

    ret = filestate._check_directory(
        root_tmp_dir,
        dir_mode=oct(expected_mode),
        file_mode=oct(expected_mode),
        recurse=["mode"],
    )
    assert changed_files == set(ret[-1].keys())



"""
    :codeauthor: `Andreas Thienemann <andreas@bawue.net>`

    tests.unit.cloud.clouds.xen_test
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import logging

import pytest

from salt.cloud.clouds import xen
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


@pytest.fixture
def configure_loader_modules():
    return {
        xen: {
            "__active_provider_name__": "",
            "__opts__": {
                "providers": {
                    "my-xen-cloud": {
                        "xen": {
                            "driver": "xen",
                            "user": "SantaClaus",
                            "password": "TooManyElves",
                            "url": "https://127.0.0.2",
                        }
                    }
                }
            },
        }
    }


def test_get_configured_provider_bad():
    with patch.dict(xen.__opts__, {"providers": {}}):
        result = xen.get_configured_provider()
        assert not result


def test_get_configured_provider_auth():
    config = {
        "url": "https://127.0.0.2",
    }
    with patch.dict(xen.__opts__, {"providers": {"my-xen-cloud": {"xen": config}}}):
        result = xen.get_configured_provider()
        assert config == result


def test_get_dependencies():
    with patch("salt.cloud.clouds.xen.HAS_XEN_API", True):
        result = xen._get_dependencies()
        assert result


def test_get_dependencies_no_xenapi():
    with patch("salt.cloud.clouds.xen.HAS_XEN_API", False):
        result = xen._get_dependencies()
        assert not result


def test_get_vm():
    XenAPI = MagicMock(name="mock_session")
    XenAPI.xenapi.VM.get_by_name_label = MagicMock(return_value=["0000"])
    XenAPI.xenapi.VM.get_is_a_template = MagicMock(return_value=False)
    with patch("salt.cloud.clouds.xen._get_session", MagicMock(return_value=XenAPI)):
        result = xen._get_vm(name="test")
        assert result == "0000"


def test_get_vm_multiple():
    """Verify correct behavior if VM and template is returned"""
    vms = {"0000": False, "0001": True}
    XenAPI = MagicMock(name="mock_session")
    XenAPI.xenapi.VM.get_by_name_label = MagicMock(return_value=vms.keys())
    XenAPI.xenapi.VM.get_is_a_template = MagicMock(side_effect=lambda x: vms[x])
    with patch("salt.cloud.clouds.xen._get_session", MagicMock(return_value=XenAPI)):
        result = xen._get_vm(name="test")
        assert result == "0000"



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import pytest

import salt.modules.monit as monit
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {monit: {}}


def test_start():
    """
    Test for start
    """
    with patch.dict(monit.__salt__, {"cmd.retcode": MagicMock(return_value=False)}):
        assert monit.start("name")


def test_stop():
    """
    Test for Stops service via monit
    """
    with patch.dict(monit.__salt__, {"cmd.retcode": MagicMock(return_value=False)}):
        assert monit.stop("name")


def test_restart():
    """
    Test for Restart service via monit
    """
    with patch.dict(monit.__salt__, {"cmd.retcode": MagicMock(return_value=False)}):
        assert monit.restart("name")


def test_unmonitor():
    """
    Test for Unmonitor service via monit
    """
    with patch.dict(monit.__salt__, {"cmd.retcode": MagicMock(return_value=False)}):
        assert monit.unmonitor("name")


def test_monitor():
    """
    Test for monitor service via monit
    """
    with patch.dict(monit.__salt__, {"cmd.retcode": MagicMock(return_value=False)}):
        assert monit.monitor("name")


def test_summary():
    """
    Test for Display a summary from monit
    """
    mock = MagicMock(side_effect=["daemon is not running", "A\nB\nC\nD\nE"])
    with patch.dict(monit.__salt__, {"cmd.run": mock}):
        assert monit.summary() == {"monit": "daemon is not running", "result": False}

        assert monit.summary() == {}


def test_status():
    """
    Test for Display a process status from monit
    """
    with patch.dict(monit.__salt__, {"cmd.run": MagicMock(return_value="Process")}):
        assert monit.status("service") == "No such service"


def test_reload():
    """
    Test for Reload configuration
    """
    mock = MagicMock(return_value=0)
    with patch.dict(monit.__salt__, {"cmd.retcode": mock}):
        assert monit.reload_()


def test_version():
    """
    Test for Display version from monit -V
    """
    mock = MagicMock(return_value="This is Monit version 5.14\nA\nB")
    with patch.dict(monit.__salt__, {"cmd.run": mock}):
        assert monit.version() == "5.14"


def test_id():
    """
    Test for Display unique id
    """
    mock = MagicMock(return_value="Monit ID: d3b1aba48527dd599db0e86f5ad97120")
    with patch.dict(monit.__salt__, {"cmd.run": mock}):
        assert monit.id_() == "d3b1aba48527dd599db0e86f5ad97120"


def test_reset_id():
    """
    Test for Regenerate a unique id
    """
    expected = {"stdout": "Monit id d3b1aba48527dd599db0e86f5ad97120 and ..."}
    mock = MagicMock(return_value=expected)
    with patch.dict(monit.__salt__, {"cmd.run_all": mock}):
        assert monit.id_(reset=True) == "d3b1aba48527dd599db0e86f5ad97120"


def test_configtest():
    """
    Test for Check configuration syntax
    """
    excepted = {"stdout": "Control file syntax OK", "retcode": 0, "stderr": ""}
    mock = MagicMock(return_value=excepted)
    with patch.dict(monit.__salt__, {"cmd.run_all": mock}):
        assert monit.configtest()["result"]
        assert monit.configtest()["comment"] == "Syntax OK"


def test_validate():
    """
    Test for Check all services are monitored
    """
    mock = MagicMock(return_value=0)
    with patch.dict(monit.__salt__, {"cmd.retcode": mock}):
        assert monit.validate()



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import urllib.error
from collections import OrderedDict

import pytest

import salt.modules.apache as apache
from tests.support.mock import MagicMock, mock_open, patch


@pytest.fixture
def configure_loader_modules():
    return {apache: {}}


# 'version' function tests: 1


def test_version():
    """
    Test if return server version (``apachectl -v``)
    """
    with patch("salt.modules.apache._detect_os", MagicMock(return_value="apachectl")):
        mock = MagicMock(return_value="Server version: Apache/2.4.7")
        with patch.dict(apache.__salt__, {"cmd.run": mock}):
            assert apache.version() == "Apache/2.4.7"


# 'fullversion' function tests: 1


def test_fullversion():
    """
    Test if return server version (``apachectl -V``)
    """
    with patch("salt.modules.apache._detect_os", MagicMock(return_value="apachectl")):
        mock = MagicMock(return_value="Server version: Apache/2.4.7")
        with patch.dict(apache.__salt__, {"cmd.run": mock}):
            assert apache.fullversion() == {
                "compiled_with": [],
                "server_version": "Apache/2.4.7",
            }


# 'modules' function tests: 1


def test_modules():
    """
    Test if return list of static and shared modules
    """
    with patch("salt.modules.apache._detect_os", MagicMock(return_value="apachectl")):
        mock = MagicMock(
            return_value=(
                "unixd_module (static)\n                             "
                " access_compat_module (shared)"
            )
        )
        with patch.dict(apache.__salt__, {"cmd.run": mock}):
            assert apache.modules() == {
                "shared": ["access_compat_module"],
                "static": ["unixd_module"],
            }


# 'servermods' function tests: 1


def test_servermods():
    """
    Test if return list of modules compiled into the server
    """
    with patch("salt.modules.apache._detect_os", MagicMock(return_value="apachectl")):
        mock = MagicMock(return_value="core.c\nmod_so.c")
        with patch.dict(apache.__salt__, {"cmd.run": mock}):
            assert apache.servermods() == ["core.c", "mod_so.c"]


# 'directives' function tests: 1


def test_directives():
    """
    Test if return list of directives
    """
    with patch("salt.modules.apache._detect_os", MagicMock(return_value="apachectl")):
        mock = MagicMock(return_value="Salt")
        with patch.dict(apache.__salt__, {"cmd.run": mock}):
            assert apache.directives() == {"Salt": ""}


# 'vhosts' function tests: 1


def test_vhosts():
    """
    Test if it shows the virtualhost settings
    """
    with patch("salt.modules.apache._detect_os", MagicMock(return_value="apachectl")):
        mock = MagicMock(return_value="")
        with patch.dict(apache.__salt__, {"cmd.run": mock}):
            assert apache.vhosts() == {}


# 'signal' function tests: 2


def test_signal():
    """
    Test if return no signal for httpd
    """
    with patch("salt.modules.apache._detect_os", MagicMock(return_value="apachectl")):
        mock = MagicMock(return_value="")
        with patch.dict(apache.__salt__, {"cmd.run": mock}):
            assert apache.signal(None) is None


def test_signal_args():
    """
    Test if return httpd signal to start, restart, or stop.
    """
    with patch("salt.modules.apache._detect_os", MagicMock(return_value="apachectl")):
        ret = 'Command: "apachectl -k start" completed successfully!'
        mock = MagicMock(return_value={"retcode": 1, "stderr": "", "stdout": ""})
        with patch.dict(apache.__salt__, {"cmd.run_all": mock}):
            assert apache.signal("start") == ret

        mock = MagicMock(
            return_value={"retcode": 1, "stderr": "Syntax OK", "stdout": ""}
        )
        with patch.dict(apache.__salt__, {"cmd.run_all": mock}):
            assert apache.signal("start") == "Syntax OK"

        mock = MagicMock(
            return_value={"retcode": 0, "stderr": "Syntax OK", "stdout": ""}
        )
        with patch.dict(apache.__salt__, {"cmd.run_all": mock}):
            assert apache.signal("start") == "Syntax OK"

        mock = MagicMock(return_value={"retcode": 1, "stderr": "", "stdout": "Salt"})
        with patch.dict(apache.__salt__, {"cmd.run_all": mock}):
            assert apache.signal("start") == "Salt"


# 'useradd' function tests: 1


def test_useradd():
    """
    Test if it add HTTP user using the ``htpasswd`` command
    """
    mock = MagicMock(return_value=True)
    with patch.dict(apache.__salt__, {"webutil.useradd": mock}):
        assert apache.useradd("htpasswd", "salt", "badpassword") is True


# 'userdel' function tests: 1


def test_userdel():
    """
    Test if it delete HTTP user using the ``htpasswd`` file
    """
    mock = MagicMock(return_value=True)
    with patch.dict(apache.__salt__, {"webutil.userdel": mock}):
        assert apache.userdel("htpasswd", "salt") is True


# 'server_status' function tests: 2


def test_server_status():
    """
    Test if return get information from the Apache server-status
    """
    with patch("salt.modules.apache.server_status", MagicMock(return_value={})):
        mock = MagicMock(return_value="")
        with patch.dict(apache.__salt__, {"config.get": mock}):
            assert apache.server_status() == {}


def test_server_status_error():
    """
    Test if return get error from the Apache server-status
    """
    mock = MagicMock(side_effect=urllib.error.URLError("error"))
    with patch("urllib.request.urlopen", mock):
        mock = MagicMock(return_value="")
        with patch.dict(apache.__salt__, {"config.get": mock}):
            assert apache.server_status() == "error"


# 'config' function tests: 1


def test_config():
    """
    Test if it create VirtualHost configuration files
    """
    with patch(
        "salt.modules.apache._parse_config", MagicMock(return_value="Listen 22")
    ):
        with patch("salt.utils.files.fopen", mock_open()):
            assert apache.config("/ports.conf", [{"Listen": "22"}]) == "Listen 22"


# '_parse_config' function tests: 2


def test__parse_config_dict():
    """
    Test parsing function which creates configs from dict like (legacy way):
        - VirtualHost:
          this: '*:80'
          ServerName: website.com
          ServerAlias:
            - www
            - dev
          Directory:
              this: /var/www/vhosts/website.com
              Order: Deny,Allow
              Allow from:
                - 127.0.0.1
                - 192.168.100.0/24

    """
    data_in = OrderedDict(
        [
            (
                "Directory",
                OrderedDict(
                    [
                        ("this", "/var/www/vhosts/website.com"),
                        ("Order", "Deny,Allow"),
                        ("Allow from", ["127.0.0.1", "192.168.100.0/24"]),
                    ]
                ),
            ),
            ("this", "*:80"),
            ("ServerName", "website.com"),
            ("ServerAlias", ["www", "dev"]),
        ]
    )
    dataout = (
        "<VirtualHost *:80>\n"
        "<Directory /var/www/vhosts/website.com>\n"
        "Order Deny,Allow\n"
        "Allow from 127.0.0.1\n"
        "Allow from 192.168.100.0/24\n\n"
        "</Directory>\n\n"
        "ServerName website.com\n"
        "ServerAlias www\n"
        "ServerAlias dev\n\n"
        "</VirtualHost>\n"
    )
    # pylint: disable=protected-access
    parse = apache._parse_config(data_in, "VirtualHost")
    assert parse == dataout


def test__parse_config_list():
    """
    Test parsing function which creates configs from variable structure (list of dicts or
    list of dicts of dicts/lists) like:
        - VirtualHost:
          - this: '*:80'
          - ServerName: website.com
          - ServerAlias:
            - www
            - dev
          - Directory:
              this: /var/www/vhosts/website.com
              Order: Deny,Allow
              Allow from:
                - 127.0.0.1
                - 192.168.100.0/24
          - Directory:
            - this: /var/www/vhosts/website.com/private
            - Order: Deny,Allow
            - Allow from:
              - 127.0.0.1
              - 192.168.100.0/24
            - If:
                this: some condition
                do: something
    """
    data_in = [
        OrderedDict(
            [
                ("ServerName", "website.com"),
                ("ServerAlias", ["www", "dev"]),
                (
                    "Directory",
                    [
                        OrderedDict(
                            [
                                ("this", "/var/www/vhosts/website.com/private"),
                                ("Order", "Deny,Allow"),
                                ("Allow from", ["127.0.0.1", "192.168.100.0/24"]),
                                (
                                    "If",
                                    {"this": "some condition", "do": "something"},
                                ),
                            ]
                        )
                    ],
                ),
                ("this", "*:80"),
            ]
        )
    ]
    dataout = (
        "<VirtualHost *:80>\n"
        "ServerName website.com\n"
        "ServerAlias www\n"
        "ServerAlias dev\n\n"
        "<Directory /var/www/vhosts/website.com/private>\n"
        "Order Deny,Allow\n"
        "Allow from 127.0.0.1\n"
        "Allow from 192.168.100.0/24\n\n"
        "<If some condition>\n"
        "do something\n"
        "</If>\n\n"
        "</Directory>\n\n"
        "</VirtualHost>\n"
    )
    # pylint: disable=protected-access
    parse = apache._parse_config(data_in, "VirtualHost")
    assert parse == dataout
